# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import time
from concurrent.futures.thread import ThreadPoolExecutor
from itertools import repeat
from typing import Any, Union

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing import Problem
from braket.aws.aws_quantum_task import AwsQuantumTask
from braket.aws.aws_session import AwsSession
from braket.circuits import Circuit
from braket.circuits.gate import Gate
from braket.ir.blackbird import Program as BlackbirdProgram
from braket.ir.openqasm import Program as OpenQasmProgram
from braket.pulse.pulse_sequence import PulseSequence
from braket.registers.qubit_set import QubitSet
from braket.tasks.quantum_task_batch import QuantumTaskBatch


class AwsQuantumTaskBatch(QuantumTaskBatch):
    """Executes a batch of quantum tasks in parallel.

    Using this class can yield vast speedups over executing quantum tasks sequentially,
    and is particularly useful for computations that can be parallelized,
    such as calculating quantum gradients or statistics of terms in a Hamiltonian.

    Note: there is no benefit to using this method with QPUs outside of their execution windows,
    since results will not be available until the window opens.
    """

    MAX_CONNECTIONS_DEFAULT = 100
    MAX_RETRIES = 3

    def __init__(
        self,
        aws_session: AwsSession,
        device_arn: str,
        task_specifications: Union[
            Union[Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation],
            list[
                Union[
                    Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation
                ]
            ],
        ],
        s3_destination_folder: AwsSession.S3DestinationFolder,
        shots: int,
        max_parallel: int,
        max_workers: int = MAX_CONNECTIONS_DEFAULT,
        poll_timeout_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        inputs: Union[dict[str, float], list[dict[str, float]]] | None = None,
        gate_definitions: (
            Union[
                dict[tuple[Gate, QubitSet], PulseSequence],
                list[dict[tuple[Gate, QubitSet], PulseSequence]],
            ]
            | None
        ) = None,
        reservation_arn: str | None = None,
        *aws_quantum_task_args: Any,
        **aws_quantum_task_kwargs: Any,
    ):
        """Creates a batch of quantum tasks.

        Args:
            aws_session (AwsSession): AwsSession to connect to AWS with.
            device_arn (str): The ARN of the quantum device.
            task_specifications (Union[Union[Circuit,Problem,OpenQasmProgram,BlackbirdProgram,AnalogHamiltonianSimulation],list[Union[Circuit,Problem,OpenQasmProgram,BlackbirdProgram,AnalogHamiltonianSimulation]]]): # noqa
                Single instance or list of circuits, annealing
                problems, pulse sequences, or photonics program as specification of quantum task
                to run on device.
            s3_destination_folder (AwsSession.S3DestinationFolder): NamedTuple, with bucket
                for index 0 and key for index 1, that specifies the Amazon S3 bucket and folder
                to store quantum task results in.
            shots (int): The number of times to run the quantum task on the device. If the device is
                a simulator, this implies the state is sampled N times, where N = `shots`.
                `shots=0` is only available on simulators and means that the simulator
                will compute the exact results based on the quantum task specification.
            max_parallel (int): The maximum number of quantum tasks to run on AWS in parallel.
                Batch creation will fail if this value is greater than the maximum allowed
                concurrent quantum tasks on the device.
            max_workers (int): The maximum number of thread pool workers. Default: 100
            poll_timeout_seconds (float): The polling timeout for `AwsQuantumTask.result()`,
                in seconds. Default: 5 days.
            poll_interval_seconds (float): The polling interval for results in seconds.
                Default: 1 second.
            inputs (Union[dict[str, float], list[dict[str, float]]] | None): Inputs to be passed
                along with the IR. If the IR supports inputs, the inputs will be updated
                with this value. Default: {}.
            gate_definitions (Union[dict[tuple[Gate, QubitSet], PulseSequence], list[dict[tuple[Gate, QubitSet], PulseSequence]]] | None): # noqa: E501
                User-defined gate calibration. The calibration is defined for a particular `Gate` on a
                particular `QubitSet` and is represented by a `PulseSequence`. Default: None.
            reservation_arn (str | None): The reservation ARN provided by Braket Direct
                to reserve exclusive usage for the device to run the quantum task on.
                Note: If you are creating tasks in a job that itself was created reservation ARN,
                those tasks do not need to be created with the reservation ARN.
                Default: None.
            *aws_quantum_task_args (Any): Arbitrary args for `QuantumTask`.
            **aws_quantum_task_kwargs (Any): Arbitrary kwargs for `QuantumTask`.,
        """  # noqa E501
        self._tasks = AwsQuantumTaskBatch._execute(
            aws_session,
            device_arn,
            task_specifications,
            s3_destination_folder,
            shots,
            max_parallel,
            max_workers,
            poll_timeout_seconds,
            poll_interval_seconds,
            inputs,
            gate_definitions,
            reservation_arn,
            *aws_quantum_task_args,
            **aws_quantum_task_kwargs,
        )
        self._aws_session = aws_session
        self._results = None
        self._unsuccessful = set()

        # Cache execution inputs for retries.
        self._device_arn = device_arn
        self._task_specifications = task_specifications
        self._s3_destination_folder = s3_destination_folder
        self._shots = shots
        self._max_parallel = max_parallel
        self._max_workers = max_workers
        self._poll_timeout_seconds = poll_timeout_seconds
        self._poll_interval_seconds = poll_interval_seconds
        self._inputs = inputs
        self._reservation_arn = reservation_arn
        self._aws_quantum_task_args = aws_quantum_task_args
        self._aws_quantum_task_kwargs = aws_quantum_task_kwargs

    @staticmethod
    def _tasks_inputs_gatedefs(
        task_specifications: Union[
            Union[Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation],
            list[
                Union[
                    Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation
                ]
            ],
        ],
        inputs: Union[dict[str, float], list[dict[str, float]]] = None,
        gate_definitions: Union[
            dict[tuple[Gate, QubitSet], PulseSequence],
            list[dict[tuple[Gate, QubitSet], PulseSequence]],
        ] = None,
    ) -> list[
        tuple[
            Union[Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation],
            dict[str, float],
            dict[tuple[Gate, QubitSet], PulseSequence],
        ]
    ]:
        inputs = inputs or {}
        gate_definitions = gate_definitions or {}

        single_task_type = (
            Circuit,
            Problem,
            OpenQasmProgram,
            BlackbirdProgram,
            AnalogHamiltonianSimulation,
        )
        single_input_type = dict
        single_gate_definitions_type = dict

        args = [task_specifications, inputs, gate_definitions]
        single_arg_types = [single_task_type, single_input_type, single_gate_definitions_type]

        batch_length = 1
        arg_lengths = []
        for arg, single_arg_type in zip(args, single_arg_types):
            arg_length = 1 if isinstance(arg, single_arg_type) else len(arg)
            arg_lengths.append(arg_length)

            if arg_length != 1:
                if batch_length != 1 and arg_length != batch_length:
                    raise ValueError(
                        "Multiple inputs, task specifications and gate definitions must "
                        "be equal in length."
                    )
                else:
                    batch_length = arg_length

        for i, arg_length in enumerate(arg_lengths):
            if isinstance(args[i], (dict, single_task_type)):
                args[i] = repeat(args[i], batch_length)

        tasks_inputs_definitions = list(zip(*args))

        for task_specification, input_map, _gate_definitions in tasks_inputs_definitions:
            if isinstance(task_specification, Circuit):
                param_names = {param.name for param in task_specification.parameters}
                unbounded_parameters = param_names - set(input_map.keys())
                if unbounded_parameters:
                    raise ValueError(
                        f"Cannot execute circuit with unbound parameters: "
                        f"{unbounded_parameters}"
                    )

        return tasks_inputs_definitions

    @staticmethod
    def _execute(
        aws_session: AwsSession,
        device_arn: str,
        task_specifications: Union[
            Union[Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation],
            list[
                Union[
                    Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation
                ]
            ],
        ],
        s3_destination_folder: AwsSession.S3DestinationFolder,
        shots: int,
        max_parallel: int,
        max_workers: int = MAX_CONNECTIONS_DEFAULT,
        poll_timeout_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        inputs: Union[dict[str, float], list[dict[str, float]]] = None,
        gate_definitions: (
            Union[
                dict[tuple[Gate, QubitSet], PulseSequence],
                list[dict[tuple[Gate, QubitSet], PulseSequence]],
            ]
            | None
        ) = None,
        reservation_arn: str | None = None,
        *args,
        **kwargs,
    ) -> list[AwsQuantumTask]:
        tasks_inputs_gatedefs = AwsQuantumTaskBatch._tasks_inputs_gatedefs(
            task_specifications, inputs, gate_definitions
        )
        max_threads = min(max_parallel, max_workers)
        remaining = [0 for _ in tasks_inputs_gatedefs]
        try:
            with ThreadPoolExecutor(max_workers=max_threads) as executor:
                task_futures = [
                    executor.submit(
                        AwsQuantumTaskBatch._create_task,
                        remaining,
                        aws_session,
                        device_arn,
                        task,
                        s3_destination_folder,
                        shots,
                        poll_timeout_seconds=poll_timeout_seconds,
                        poll_interval_seconds=poll_interval_seconds,
                        inputs=input_map,
                        gate_definitions=gatedefs,
                        reservation_arn=reservation_arn,
                        *args,
                        **kwargs,
                    )
                    for task, input_map, gatedefs in tasks_inputs_gatedefs
                ]
        except KeyboardInterrupt:
            # If an exception is thrown before the thread pool has finished,
            # clean up the quantum tasks which have not yet been created before reraising it.
            if "task_futures" in locals():
                for future in task_futures:
                    future.cancel()

            # Signal to the workers that there is no mork work to do
            remaining.clear()

            raise
        tasks = [future.result() for future in task_futures]
        return tasks

    @staticmethod
    def _create_task(
        remaining: list[int],
        aws_session: AwsSession,
        device_arn: str,
        task_specification: Union[
            Circuit, Problem, OpenQasmProgram, BlackbirdProgram, AnalogHamiltonianSimulation
        ],
        s3_destination_folder: AwsSession.S3DestinationFolder,
        shots: int,
        poll_interval_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL,
        inputs: dict[str, float] = None,
        gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence] | None = None,
        reservation_arn: str | None = None,
        *args,
        **kwargs,
    ) -> AwsQuantumTask:
        task = AwsQuantumTask.create(
            aws_session,
            device_arn,
            task_specification,
            s3_destination_folder,
            shots,
            poll_interval_seconds=poll_interval_seconds,
            inputs=inputs,
            gate_definitions=gate_definitions,
            reservation_arn=reservation_arn,
            *args,
            **kwargs,
        )

        remaining.pop()

        # If the quantum task hits a terminal state before all quantum tasks have been created,
        # it can be returned immediately
        while remaining:
            if task.state() in AwsQuantumTask.TERMINAL_STATES:
                break
            time.sleep(poll_interval_seconds)
        return task

    def results(
        self,
        fail_unsuccessful: bool = False,
        max_retries: int = MAX_RETRIES,
        use_cached_value: bool = True,
    ) -> list[AwsQuantumTask]:
        """Retrieves the result of every quantum task in the batch.

        Polling for results happens in parallel; this method returns when all quantum tasks
        have reached a terminal state. The result of this method is cached.

        Args:
            fail_unsuccessful (bool): If set to `True`, this method will fail
                if any quantum task in the batch fails to return a result even after
                `max_retries` retries.
            max_retries (int): Maximum number of times to retry any failed quantum tasks,
                i.e. any quantum tasks in the `FAILED` or `CANCELLED` state or that didn't
                complete within the timeout. Default: 3.
            use_cached_value (bool): If `False`, will refetch the results from S3,
                even when results have already been cached. Default: `True`.

        Returns:
            list[AwsQuantumTask]: The results of all of the quantum tasks in the batch.
            `FAILED`, `CANCELLED`, or timed out quantum tasks will have a result of None
        """
        if not self._results or not use_cached_value:
            self._results = AwsQuantumTaskBatch._retrieve_results(self._tasks, self._max_workers)
            self._unsuccessful = {
                task.id for task, result in zip(self._tasks, self._results) if not result
            }

        retries = 0
        while self._unsuccessful and retries < max_retries:
            self.retry_unsuccessful_tasks()
            retries = retries + 1

        if fail_unsuccessful and self._unsuccessful:
            raise RuntimeError(
                f"{len(self._unsuccessful)} tasks failed to complete after {max_retries} retries"
            )
        return self._results

    @staticmethod
    def _retrieve_results(tasks: list[AwsQuantumTask], max_workers: int) -> list[AwsQuantumTask]:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            result_futures = [executor.submit(task.result) for task in tasks]
        return [future.result() for future in result_futures]

    def retry_unsuccessful_tasks(self) -> bool:
        """Retries any quantum tasks in the batch without valid results.

        This method should only be called after `results()` has been called at least once.
        The method will generate new quantum tasks for any failed quantum tasks, so `self.task` and
        `self.results()` may return different values after a call to this method.

        Returns:
            bool: Whether or not all retried quantum tasks completed successfully.
        """
        if not self._results:
            raise RuntimeError("results() should be called before attempting to retry")
        unsuccessful_indices = [index for index, result in enumerate(self._results) if not result]
        # Return early if there's nothing to retry
        if not unsuccessful_indices:
            return True
        retried_tasks = AwsQuantumTaskBatch._execute(
            self._aws_session,
            self._device_arn,
            [self._task_specifications[i] for i in unsuccessful_indices],
            self._s3_destination_folder,
            self._shots,
            self._max_parallel,
            self._max_workers,
            self._poll_timeout_seconds,
            self._poll_interval_seconds,
            self._reservation_arn,
            *self._aws_quantum_task_args,
            **self._aws_quantum_task_kwargs,
        )
        for index, task in zip(unsuccessful_indices, retried_tasks):
            self._tasks[index] = task

        retried_results = AwsQuantumTaskBatch._retrieve_results(retried_tasks, self._max_workers)
        for index, result in zip(unsuccessful_indices, retried_results):
            self._results[index] = result
        self._unsuccessful = {
            task.id for task, result in zip(retried_tasks, retried_results) if not result
        }
        return not self._unsuccessful

    @property
    def tasks(self) -> list[AwsQuantumTask]:
        """list[AwsQuantumTask]: The quantum tasks in this batch, as a list of AwsQuantumTask
        objects
        """
        return list(self._tasks)

    @property
    def size(self) -> int:
        """int: The number of quantum tasks in the batch"""
        return len(self._tasks)

    @property
    def unfinished(self) -> set[str]:
        """Gets all the IDs of all the quantum tasks in teh batch that have yet to complete.

        Returns:
            set[str]: The IDs of all the quantum tasks in the batch that have yet to complete.
        """
        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            status_futures = {task.id: executor.submit(task.state) for task in self._tasks}
        unfinished = set()
        for task_id in status_futures:
            status = status_futures[task_id].result()
            if status not in AwsQuantumTask.TERMINAL_STATES:
                unfinished.add(task_id)
            if status in AwsQuantumTask.NO_RESULT_TERMINAL_STATES:
                self._unsuccessful.add(task_id)
        return unfinished

    @property
    def unsuccessful(self) -> set[str]:
        """set[str]: The IDs of all the FAILED, CANCELLED, or timed out quantum tasks in the
        batch.
        """
        return set(self._unsuccessful)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.jobs.image_uris import Framework, retrieve_image


@pytest.mark.parametrize(
    "region, framework, expected_uri",
    [
        (
            "us-west-1",
            Framework.BASE,
            "292282985366.dkr.ecr.us-west-1.amazonaws.com/amazon-braket-base-jobs:latest",
        ),
        (
            "us-east-1",
            Framework.PL_TENSORFLOW,
            "292282985366.dkr.ecr.us-east-1.amazonaws.com/amazon-braket-tensorflow-jobs:latest",
        ),
        (
            "us-west-2",
            Framework.PL_PYTORCH,
            "292282985366.dkr.ecr.us-west-2.amazonaws.com/amazon-braket-pytorch-jobs:latest",
        ),
    ],
)
def test_retrieve_image_default_version(region, framework, expected_uri):
    assert retrieve_image(framework, region) == expected_uri


@pytest.mark.parametrize(
    "region, framework",
    [
        ("eu-west-1", Framework.BASE),
        (None, Framework.BASE),
        ("us-west-1", None),
        ("us-west-1", "foo"),
    ],
)
@pytest.mark.xfail(raises=ValueError)
def test_retrieve_image_incorrect_input(region, framework):
    retrieve_image(framework, region)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.circuits import Gate, Instruction, Noise, Qubit, QubitSet, compiler_directives
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)


@pytest.fixture
def instr():
    return Instruction(Gate.H(), 0)


@pytest.fixture
def cnot():
    return Instruction(Gate.CNot(), [0, 1])


@pytest.fixture
def ccry():
    return Instruction(Gate.Ry(1.23), 0, control=[1, 2])


def test_empty_operator():
    with pytest.raises(ValueError):
        Instruction(None, target=0)


def test_non_matching_qubit_set_and_qubit_count():
    with pytest.raises(ValueError):
        Instruction(Gate.CNot(), target=[0, 0])


def test_init_with_qubits():
    target = QubitSet([0, 1])
    instr = Instruction(Gate.CNot(), target)
    assert instr.target == target


def test_init_with_qubit():
    target = Qubit(0)
    instr = Instruction(Gate.H(), target)
    assert instr.target == QubitSet(0)


def test_init_with_int():
    target = 0
    instr = Instruction(Gate.H(), target)
    assert instr.target == QubitSet(0)


def test_init_with_sequence():
    target = [0, Qubit(1)]
    instr = Instruction(Gate.CNot(), target)
    assert instr.target == QubitSet([0, 1])


def test_getters():
    target = [0, 1]
    operator = Gate.CNot()
    instr = Instruction(operator, target)

    assert instr.operator == operator
    assert instr.target == QubitSet([0, 1])


def test_operator_setter(instr):
    with pytest.raises(AttributeError):
        instr.operator = Gate.H()


def test_target_setter(instr):
    with pytest.raises(AttributeError):
        instr.target = QubitSet(0)


def test_adjoint_gate():
    instr = Instruction(Gate.S(), 0)
    adj = instr.adjoint()
    assert adj == [Instruction(Gate.Si(), 0)]


def test_adjoint_compiler_directive():
    instr = Instruction(compiler_directives.StartVerbatimBox()).adjoint()
    assert instr == [Instruction(compiler_directives.EndVerbatimBox())]


def test_adjoint_unsupported():
    with pytest.raises(NotImplementedError):
        Instruction(Noise.BitFlip(0.1), 0).adjoint()


def test_str(instr):
    expected = (
        "Instruction('operator': {}, 'target': {}, "
        "'control': {}, 'control_state': {}, 'power': {})"
    ).format(
        instr.operator,
        instr.target,
        instr.control,
        instr.control_state.as_tuple,
        instr.power,
    )
    assert str(instr) == expected


def test_equality():
    instr_1 = Instruction(Gate.H(), 0)
    instr_2 = Instruction(Gate.H(), 0)
    other_instr = Instruction(Gate.CNot(), [0, 1])
    non_instr = "non instruction"

    assert instr_1 == instr_2
    assert instr_1 is not instr_2
    assert instr_1 != other_instr
    assert instr_1 != non_instr


def test_to_ir():
    expected_target = QubitSet([0, 1])
    expected_ir = "foo bar value"
    expected_ir_type = IRType.OPENQASM
    expected_serialization_properties = OpenQASMSerializationProperties(
        qubit_reference_type=QubitReferenceType.PHYSICAL
    )

    class FooGate(Gate):
        def __init__(self):
            super().__init__(qubit_count=2, ascii_symbols=["foo", "bar"])

        def to_ir(self, target, ir_type, serialization_properties):
            assert target == expected_target
            assert ir_type == expected_ir_type
            assert serialization_properties == expected_serialization_properties
            return expected_ir

    instr = Instruction(FooGate(), expected_target)
    assert instr.to_ir(expected_ir_type, expected_serialization_properties) == expected_ir


def test_copy_creates_new_object(instr):
    copy = instr.copy()
    assert copy == instr
    assert copy is not instr


def test_copy_with_mapping(cnot):
    target_mapping = {0: 10, 1: 11}
    expected = Instruction(Gate.CNot(), [10, 11])
    assert cnot.copy(target_mapping=target_mapping) == expected


def test_copy_with_control_mapping(ccry):
    control_mapping = {1: 10, 2: 11}
    expected = Instruction(Gate.Ry(1.23), target=1, control=[10, 11])
    assert ccry.copy(target=1, control_mapping=control_mapping) == expected


def test_copy_with_target(cnot):
    target = [10, 11]
    expected = Instruction(Gate.CNot(), target)
    assert cnot.copy(target=target) == expected


def test_copy_with_control(ccry):
    control = [10, 11]
    expected = Instruction(Gate.Ry(1.23), 3, control=control)
    assert ccry.copy(target=3, control=control) == expected


def test_copy_with_target_and_mapping(instr):
    cant_do_both = "Only 'target_mapping' or 'target' can be supplied, but not both."
    with pytest.raises(TypeError, match=cant_do_both):
        instr.copy(target=[10], target_mapping={0: 10})


def test_copy_with_control_target_and_mapping(instr):
    cant_do_both = "Only 'control_mapping' or 'control' can be supplied, but not both."
    with pytest.raises(TypeError, match=cant_do_both):
        instr.copy(target=[10], control=[10], control_mapping={0: 10})


def test_pow(instr):
    assert instr.power == 1
    cubed = instr**3
    assert instr.power == 1
    assert cubed.power == 3
    then_squared = cubed**2
    assert cubed.power == 3
    assert then_squared.power == 6
    modded = then_squared.__pow__(6, 5)
    assert modded.power == 1



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest
from gate_model_device_testing_utils import (
    multithreaded_bell_pair_testing,
    no_result_types_bell_pair_testing,
    qubit_ordering_testing,
    result_types_all_selected_testing,
    result_types_bell_pair_full_probability_testing,
    result_types_bell_pair_marginal_probability_testing,
    result_types_hermitian_testing,
    result_types_noncommuting_all,
    result_types_noncommuting_flipped_targets_testing,
    result_types_noncommuting_testing,
    result_types_nonzero_shots_bell_pair_testing,
    result_types_observable_not_in_instructions,
    result_types_tensor_hermitian_hermitian_testing,
    result_types_tensor_x_y_testing,
    result_types_tensor_y_hermitian_testing,
    result_types_tensor_z_h_y_testing,
    result_types_tensor_z_hermitian_testing,
    result_types_tensor_z_z_testing,
    result_types_zero_shots_bell_pair_testing,
)

from braket.devices import LocalSimulator

DEVICE = LocalSimulator()
SHOTS = 8000


def test_multithreaded_bell_pair(caplog):
    multithreaded_bell_pair_testing(DEVICE, {"shots": SHOTS})
    assert not caplog.text


def test_no_result_types_bell_pair(caplog):
    no_result_types_bell_pair_testing(DEVICE, {"shots": SHOTS})
    assert not caplog.text


def test_qubit_ordering(caplog):
    qubit_ordering_testing(DEVICE, {"shots": SHOTS})
    assert not caplog.text


def test_result_types_no_shots(caplog):
    result_types_zero_shots_bell_pair_testing(DEVICE, True, {"shots": 0})
    assert not caplog.text


def test_result_types_nonzero_shots_bell_pair(caplog):
    result_types_nonzero_shots_bell_pair_testing(DEVICE, {"shots": SHOTS})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_bell_pair_full_probability(shots, caplog):
    result_types_bell_pair_full_probability_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_bell_pair_marginal_probability(shots, caplog):
    result_types_bell_pair_marginal_probability_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_hermitian(shots, caplog):
    result_types_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_x_y(shots, caplog):
    result_types_tensor_x_y_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_z_z(shots, caplog):
    result_types_tensor_z_z_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_z_h_y(shots, caplog):
    result_types_tensor_z_h_y_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_z_hermitian(shots, caplog):
    result_types_tensor_z_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_hermitian_hermitian(shots, caplog):
    result_types_tensor_hermitian_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_y_hermitian(shots, caplog):
    result_types_tensor_y_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_all_selected(shots, caplog):
    result_types_all_selected_testing(DEVICE, {"shots": shots})
    assert not caplog.text


def test_result_types_noncommuting(caplog):
    result_types_noncommuting_testing(DEVICE, {})
    assert not caplog.text


def test_result_types_noncommuting_flipped_targets(caplog):
    result_types_noncommuting_flipped_targets_testing(DEVICE, {})
    assert not caplog.text


def test_result_types_noncommuting_all(caplog):
    result_types_noncommuting_all(DEVICE, {})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_observable_not_in_instructions(shots, caplog):
    result_types_observable_not_in_instructions(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize(
    "backend, device_name",
    [
        ("default", "StateVectorSimulator"),
        ("braket_sv", "StateVectorSimulator"),
        ("braket_dm", "DensityMatrixSimulator"),
        ("braket_ahs", "RydbergAtomSimulator"),
    ],
)
def test_local_simulator_device_names(backend, device_name, caplog):
    local_simulator_device = LocalSimulator(backend)
    assert local_simulator_device.name == device_name
    assert not caplog.text



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re
from collections.abc import Iterator
from logging import Logger, getLogger
from typing import Optional, Union

from braket.jobs.metrics_data.definitions import MetricStatistic, MetricType


class LogMetricsParser:
    """This class is used to parse metrics from log lines, and return them in a more
    convenient format.
    """

    METRICS_DEFINITIONS = re.compile(r"(\w+)\s*=\s*([^;]+)\s*;")
    TIMESTAMP = "timestamp"
    ITERATION_NUMBER = "iteration_number"
    NODE_ID = "node_id"
    NODE_TAG = re.compile(r"^\[([^\]]*)\]")

    def __init__(
        self,
        logger: Logger = getLogger(__name__),
    ):
        self._logger = logger
        self.all_metrics = []

    @staticmethod
    def _get_value(
        current_value: Optional[Union[str, float, int]],
        new_value: Union[str, float, int],
        statistic: MetricStatistic,
    ) -> Union[str, float, int]:
        """Gets the value based on a statistic.

        Args:
            current_value (Optional[Union[str, float, int]]): The current value.

            new_value (Union[str, float, int]): The new value.

            statistic (MetricStatistic): The statistic to determine which value to use.

        Returns:
            Union[str, float, int]: the value.
        """
        if current_value is None:
            return new_value
        if statistic == MetricStatistic.MAX:
            return max(current_value, new_value)
        return min(current_value, new_value)

    def _get_metrics_from_log_line_matches(
        self, all_matches: Iterator
    ) -> dict[str, Union[str, float, int]]:
        """Converts matches from a RegEx to a set of metrics.

        Args:
            all_matches (Iterator): An iterator for RegEx matches on a log line.

        Returns:
            dict[str, Union[str, float, int]]: The set of metrics found by the RegEx. The result
            is in the format {<metric name> : <value>}. This implies that multiple metrics
            with the same name are deduped to the last instance of that metric.
        """
        metrics = {}
        for match in all_matches:
            subgroup = match.groups()
            value = subgroup[1]
            try:
                metrics[subgroup[0]] = float(value)
            except ValueError:
                self._logger.warning(f"Unable to convert value {value} to a float.")
        return metrics

    def parse_log_message(self, timestamp: str, message: str) -> None:
        """Parses a line from logs, adding all the metrics that have been logged
        on that line. The timestamp is also added to match the corresponding values.

        Args:
            timestamp (str): A formatted string representing the timestamp for any found metrics.

            message (str): A log line from a log.
        """
        if not message:
            return
        all_matches = self.METRICS_DEFINITIONS.finditer(message)
        parsed_metrics = self._get_metrics_from_log_line_matches(all_matches)
        if not parsed_metrics:
            return
        if timestamp and self.TIMESTAMP not in parsed_metrics:
            parsed_metrics[self.TIMESTAMP] = timestamp
        node_match = self.NODE_TAG.match(message)
        if node_match:
            parsed_metrics[self.NODE_ID] = node_match.group(1)
        self.all_metrics.append(parsed_metrics)

    def get_columns_and_pivot_indices(
        self, pivot: str
    ) -> tuple[dict[str, list[Union[str, float, int]]], dict[tuple[int, str], int]]:
        """Parses the metrics to find all the metrics that have the pivot column. The values of the
        pivot column are paired with the node_id and assigned a row index, so that all metrics
        with the same pivot value and node_id are stored in the same row.

        Args:
            pivot (str): The name of the pivot column. Must be TIMESTAMP or ITERATION_NUMBER.

        Returns:
            tuple[dict[str, list[Union[str, float, int]]], dict[tuple[int, str], int]]: Contains:
            The dict[str, list[Any]] is the result table with all the metrics values initialized
            to None.
            The dict[tuple[int, str], int] is the list of pivot indices, where the value of a
            pivot column and node_id is mapped to a row index.
        """
        row_count = 0
        pivot_indices: dict[int, int] = {}
        table: dict[str, list[Optional[Union[str, float, int]]]] = {}
        for metric in self.all_metrics:
            if pivot in metric:
                # If no node_id is present, pair pivot value with None for the key.
                metric_pivot = (metric[pivot], metric.get(self.NODE_ID))
                if metric_pivot not in pivot_indices:
                    pivot_indices[metric_pivot] = row_count
                    row_count += 1
                for column_name in metric:
                    table[column_name] = [None]
        for column_name in table:
            table[column_name] = [None] * row_count
        return table, pivot_indices

    def get_metric_data_with_pivot(
        self, pivot: str, statistic: MetricStatistic
    ) -> dict[str, list[Union[str, float, int]]]:
        """Gets the metric data for a given pivot column name. Metrics without the pivot column
        are not included in the results. Metrics that have the same value in the pivot column
        from the same node are returned in the same row. Metrics from different nodes are stored
        in different rows. If the metric has multiple values for the row, the statistic is used
        to determine which value is returned.
        For example, for the metrics:
        "iteration_number" : 0, "metricA" : 2, "metricB" : 1,
        "iteration_number" : 0, "metricA" : 1,
        "no_pivot_column" : 0,  "metricA" : 0,
        "iteration_number" : 1, "metricA" : 2,
        "iteration_number" : 1, "node_id" : "nodeB", "metricB" : 0,

        The result with iteration_number as the pivot, statistic of MIN the result will be:
            iteration_number node_id metricA metricB
            0                None    1       1
            1                None    2       None
            1                nodeB   None    0

        Args:
            pivot (str): The name of the pivot column. Must be TIMESTAMP or ITERATION_NUMBER.
            statistic (MetricStatistic): The statistic to determine which value to use.

        Returns:
            dict[str, list[Union[str, float, int]]]: The metrics data.
        """
        table, pivot_indices = self.get_columns_and_pivot_indices(pivot)
        for metric in self.all_metrics:
            if pivot in metric:
                metric_pivot = (metric[pivot], metric.get(self.NODE_ID))
                row = pivot_indices[metric_pivot]
                for column_name in metric:
                    table[column_name][row] = self._get_value(
                        table[column_name][row], metric[column_name], statistic
                    )
        return table

    def get_parsed_metrics(
        self, metric_type: MetricType, statistic: MetricStatistic
    ) -> dict[str, list[Union[str, float, int]]]:
        """Gets all the metrics data, where the keys are the column names and the values are a list
        containing the values in each row.

        Args:
            metric_type (MetricType): The type of metrics to get.

            statistic (MetricStatistic): The statistic to determine which metric value to use
                when there is a conflict.

        Returns:
            dict[str, list[Union[str, float, int]]]: The metrics data.

        Example:
            timestamp energy
              0         0.1
              1         0.2
            would be represented as:
            { "timestamp" : [0, 1], "energy" : [0.1, 0.2] }
            values may be integers, floats, strings or None.
        """
        if metric_type == MetricType.ITERATION_NUMBER:
            return self.get_metric_data_with_pivot(self.ITERATION_NUMBER, statistic)
        return self.get_metric_data_with_pivot(self.TIMESTAMP, statistic)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import Optional

from braket.circuits.circuit import Circuit
from braket.circuits.gate import Gate
from braket.circuits.instruction import Instruction
from braket.circuits.noise import Noise
from braket.circuits.noise_model.circuit_instruction_criteria import CircuitInstructionCriteria
from braket.circuits.noise_model.criteria import Criteria, CriteriaKey, CriteriaKeyResult
from braket.circuits.noise_model.initialization_criteria import InitializationCriteria
from braket.circuits.noise_model.result_type_criteria import ResultTypeCriteria
from braket.circuits.result_types import ObservableResultType
from braket.registers.qubit_set import QubitSetInput


@dataclass
class NoiseModelInstruction:
    """Represents a single instruction for a Noise Model."""

    noise: Noise
    criteria: Criteria

    def __init__(self, noise: Noise, criteria: Criteria):
        if not isinstance(noise, Noise):
            raise ValueError(f"{noise} must be a Noise type.")
        if not isinstance(criteria, Criteria):
            raise ValueError(f"{criteria} must be a Criteria type.")
        self.noise = noise
        self.criteria = criteria

    def __repr__(self):
        return f"NoiseModelInstruction(noise={self.noise}, criteria={self.criteria})"

    def __str__(self):
        return f"{self.noise}, {self.criteria}"

    def to_dict(self) -> dict:
        """Converts this object to a dictionary."""
        return {"noise": self.noise.to_dict(), "criteria": self.criteria.to_dict()}

    @classmethod
    def from_dict(cls, noise_model_item: dict) -> NoiseModelInstruction:
        """Converts a dictionary representing an object of this class into an instance of
        this class.

        Args:
            noise_model_item (dict): A dictionary representation of an object of this class.

        Returns:
            NoiseModelInstruction: An object of this class that corresponds
            to the passed in dictionary.
        """
        return NoiseModelInstruction(
            noise=Noise.from_dict(noise_model_item["noise"]),
            criteria=Criteria.from_dict(noise_model_item["criteria"]),
        )


@dataclass
class NoiseModelInstructions:
    """Represents the instructions in a noise model, separated by type."""

    initialization_noise: list[NoiseModelInstruction]
    gate_noise: list[NoiseModelInstruction]
    readout_noise: list[NoiseModelInstruction]


class NoiseModel:
    """A Noise Model can be thought of as a set of Noise objects, and a corresponding set of
    criteria for how each Noise object should be applied to a circuit. For example, a noise model
    may represent that every H gate that acts on qubit 0 has a 10% probability of a bit flip, and
    every X gate that acts on qubit 0 has a 20% probability of a bit flip, and 5% probability of
    a phase flip.
    """

    def __init__(self, instructions: list[NoiseModelInstruction] = None):
        self._instructions = instructions or []

    def __repr__(self):
        return str({"instructions": list(self._instructions)})

    def __str__(self):
        instructions = self.get_instructions_by_type()
        all_strings = []
        all_strings.extend(
            NoiseModel._items_to_string("Initialization Noise:", instructions.initialization_noise)
        )
        all_strings.extend(NoiseModel._items_to_string("Gate Noise:", instructions.gate_noise))
        all_strings.extend(
            NoiseModel._items_to_string("Readout Noise:", instructions.readout_noise)
        )
        return "\n".join(all_strings)

    @property
    def instructions(self) -> list[NoiseModelInstruction]:
        """List all the noise in the NoiseModel.

        Returns:
            list[NoiseModelInstruction]: The noise model instructions.
        """
        return self._instructions

    def add_noise(self, noise: Noise, criteria: Criteria) -> NoiseModel:
        """Modifies the NoiseModel to add noise with a given Criteria.

        Args:
            noise (Noise): The noise to add.
            criteria (Criteria): The criteria that determines when the noise will be applied.

        Returns:
            NoiseModel: This NoiseModel object.
        """
        return self._add_instruction(NoiseModelInstruction(noise, criteria))

    def insert_noise(self, index: int, noise: Noise, criteria: Criteria) -> NoiseModel:
        """Modifies the NoiseModel to insert a noise with a given Criteria at particular position.

        Args:
            index (int): The index at which to insert.
            noise (Noise): The noise to insert.
            criteria (Criteria): The criteria that determines when the noise will be applied.

        Returns:
            NoiseModel: This NoiseModel object.
        """
        self._instructions.insert(index, NoiseModelInstruction(noise, criteria))
        return self

    def _add_instruction(self, instruction: NoiseModelInstruction) -> NoiseModel:
        """Modifies the NoiseModel to add noise with a given Criteria.

        Args:
            instruction (NoiseModelInstruction): The noise model instruction to add.

        Returns:
            NoiseModel: This NoiseModel object.
        """
        self._instructions.append(instruction)
        return self

    def remove_noise(self, index: int) -> NoiseModel:
        """Removes the noise and corresponding criteria from the NoiseModel at the given index.

        Args:
            index (int): The index of the instruction to remove.

        Returns:
            NoiseModel: This NoiseModel object.

        Throws:
            IndexError: If the provided index is not less than the number of noise (as returned
                from items()).
        """
        self._instructions.pop(index)
        return self

    def get_instructions_by_type(self) -> NoiseModelInstructions:
        """Returns the noise in this model by noise type.

        Returns:
            NoiseModelInstructions: The noise model instructions.
        """
        init_noise = []
        gate_noise = []
        readout_noise = []
        for item in self._instructions:
            if isinstance(item.criteria, InitializationCriteria):
                init_noise.append(item)
            elif isinstance(item.criteria, CircuitInstructionCriteria):
                gate_noise.append(item)
            elif isinstance(item.criteria, ResultTypeCriteria):
                readout_noise.append(item)
        return NoiseModelInstructions(
            initialization_noise=init_noise,
            gate_noise=gate_noise,
            readout_noise=readout_noise,
        )

    def from_filter(
        self,
        qubit: Optional[QubitSetInput] = None,
        gate: Optional[Gate] = None,
        noise: Optional[type[Noise]] = None,
    ) -> NoiseModel:
        """Returns a new NoiseModel from this NoiseModel using a given filter. If no filters are
        specified, the returned NoiseModel will be the same as this one.

        Args:
            qubit (Optional[QubitSetInput]): The qubit to filter. Default is None.
                If not None, the returned NoiseModel will only have Noise that might be applicable
                to the passed qubit (or qubit list, if the criteria acts on a multi-qubit Gate).
            gate (Optional[Gate]): The gate to filter. Default is None. If not None,
                the returned NoiseModel will only have Noise that might be applicable
                to the passed Gate.
            noise (Optional[type[Noise]]): The noise class to filter. Default is None.
                If not None, the returned NoiseModel will only have noise that is of the same
                class as the given noise class.

        Returns:
            NoiseModel: A noise model containing Noise and Criteria that are applicable for
            the given filter.
        """
        new_model = NoiseModel()
        for item in self._instructions:
            if noise is not None and not isinstance(item.noise, noise):
                continue
            if gate is not None:
                gate_keys = item.criteria.get_keys(CriteriaKey.GATE)
                if gate_keys != CriteriaKeyResult.ALL and gate not in gate_keys:
                    continue
            if qubit is not None:
                qubit_keys = item.criteria.get_keys(CriteriaKey.QUBIT)
                if qubit_keys != CriteriaKeyResult.ALL and qubit not in qubit_keys:
                    continue
            new_model.add_noise(item.noise, item.criteria)
        return new_model

    def apply(self, circuit: Circuit) -> Circuit:
        """Applies this noise model to a circuit, and returns a new circuit that's the `noisy`
        version of the given circuit. If multiple noise will act on the same instruction,
        they will be applied in the order they are added to the noise model.

        Args:
            circuit (Circuit): a circuit to apply `noise` to.

        Returns:
            Circuit: A new circuit that's a `noisy` version of the passed in circuit.
        """
        instructions = self.get_instructions_by_type()
        new_circuit = NoiseModel._apply_gate_noise(circuit, instructions.gate_noise)
        new_circuit = NoiseModel._apply_init_noise(new_circuit, instructions.initialization_noise)
        return NoiseModel._apply_readout_noise(new_circuit, instructions.readout_noise)

    def to_dict(self) -> dict:
        """Converts this object to a dictionary."""
        return {"instructions": [item.to_dict() for item in self._instructions]}

    @classmethod
    def _apply_gate_noise(
        cls,
        circuit: Circuit,
        gate_noise_instructions: list[NoiseModelInstruction],
    ) -> Circuit:
        """Applies the gate noise to return a new circuit that's the `noisy` version of the given
        circuit.

        Args:
            circuit (Circuit): a circuit to apply `noise` to.
            gate_noise_instructions (list[NoiseModelInstruction]): a list of gate noise
                instructions to apply to the circuit.

        Returns:
            Circuit: A new circuit that's a `noisy` version of the passed in circuit. The targets
            set will be populated with the list of targets in the new circuit.
        """
        new_circuit = Circuit()
        for circuit_instruction in circuit.instructions:
            new_circuit.add_instruction(circuit_instruction)
            target_qubits = list(circuit_instruction.target)
            for item in gate_noise_instructions:
                if item.criteria.instruction_matches(circuit_instruction):
                    if item.noise.fixed_qubit_count() == len(target_qubits):
                        new_circuit.add_instruction(Instruction(item.noise, target_qubits))
                    else:
                        for qubit in target_qubits:
                            new_circuit.add_instruction(Instruction(item.noise, qubit))
        for result_type in circuit.result_types:
            new_circuit.add_result_type(result_type)
        return new_circuit

    @classmethod
    def _apply_init_noise(
        cls,
        circuit: Circuit,
        init_noise_instructions: list[NoiseModelInstruction],
    ) -> Circuit:
        """Applies the initialization noise of this noise model to a circuit and returns
        the circuit.

        Args:
            circuit (Circuit): A circuit to apply `noise` to.
            init_noise_instructions (list[NoiseModelInstruction]): A list of initialization noise
                model instructions.

        Returns:
            Circuit: The passed in circuit, with the initialization noise applied.
        """
        if not init_noise_instructions:
            return circuit
        for item in init_noise_instructions:
            qubits = item.criteria.qubit_intersection(circuit.qubits)
            if len(qubits) > 0:
                circuit.apply_initialization_noise(item.noise, list(qubits))
        return circuit

    @classmethod
    def _apply_readout_noise(
        cls,
        circuit: Circuit,
        readout_noise_instructions: list[NoiseModelInstruction],
    ) -> Circuit:
        """Applies the readout noise of this noise model to a circuit and returns the circuit.

        Args:
            circuit (Circuit): A circuit to apply `noise` to.
            readout_noise_instructions (list[NoiseModelInstruction]): The list of readout noise
                to apply.

        Returns:
            Circuit: The passed in circuit, with the readout noise applied.
        """
        if not readout_noise_instructions or not circuit.result_types:
            return circuit
        return _apply_noise_on_observable_result_types(circuit, readout_noise_instructions)

    @classmethod
    def _items_to_string(
        cls, instructions_title: str, instructions: list[NoiseModelInstruction]
    ) -> list[str]:
        """Creates a string representation of a list of instructions.

        Args:
            instructions_title (str): The title for this list of instructions.
            instructions (list[NoiseModelInstruction]): A list of instructions.

        Returns:
            list[str]: A list of string representations of the passed instructions.
        """
        results = []
        if len(instructions) > 0:
            results.append(instructions_title)
            for item in instructions:
                results.append(f"  {item}")
        return results

    @classmethod
    def from_dict(cls, noise_dict: dict) -> NoiseModel:
        """Converts a dictionary representing an object of this class into an instance
        of this class.

        Args:
            noise_dict (dict): A dictionary representation of an object of this class.

        Returns:
            NoiseModel: An object of this class that corresponds to the passed in dictionary.
        """
        model = NoiseModel()
        all(
            model._add_instruction(NoiseModelInstruction.from_dict(item))
            for item in noise_dict["instructions"]
        )
        return model


def _apply_noise_on_observable_result_types(
    circuit: Circuit, readout_noise_instructions: list[NoiseModelInstruction]
) -> Circuit:
    """Applies readout noise based on the observable result types in the circuit. Each applicable
    Noise will be applied only once to a target in the ObservableResultType.

    Args:
        circuit (Circuit): The circuit to apply the readout noise to.
        readout_noise_instructions (list[NoiseModelInstruction]): The list of readout noise
            to apply.

    Returns:
        Circuit: The passed in circuit, with the readout noise applied.
    """
    result_types = circuit.result_types
    noise_to_apply = defaultdict(set)
    for result_type in result_types:
        if isinstance(result_type, ObservableResultType):
            target_qubits = list(result_type.target)
            for item_index, item in enumerate(readout_noise_instructions):
                if item.criteria.result_type_matches(result_type):
                    for target_qubit in target_qubits:
                        noise_to_apply[target_qubit].add(item_index)
    for qubit in noise_to_apply:
        for noise_item_index in noise_to_apply[qubit]:
            item = readout_noise_instructions[noise_item_index]
            circuit.apply_readout_noise(item.noise, qubit)
    return circuit



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from dataclasses import dataclass

from braket.timings.time_series import TimeSeries


@dataclass
class PulseSequenceTrace:
    """This class encapsulates the data representing the PulseSequence execution. It contains
    the trace of amplitude, frequency and phase information for each frame in the
    PulseSequence.

    Args:
        amplitudes (dict): A dictionary of frame ID to a TimeSeries of complex values specifying
            the waveform amplitude.
        frequencies (dict):A dictionary of frame ID to a TimeSeries of float values specifying
            the waveform frequency.
        phases (dict):A dictionary of frame ID to a TimeSeries of float values specifying
            the waveform phase.
    """

    amplitudes: dict[str, TimeSeries]
    frequencies: dict[str, TimeSeries]
    phases: dict[str, TimeSeries]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re
from collections import defaultdict
from collections.abc import KeysView
from dataclasses import dataclass
from typing import Any, ClassVar, Optional, Union

import numpy as np
from openpulse import ast
from openqasm3.visitor import QASMVisitor
from oqpy import Program

from braket.pulse.frame import Frame
from braket.pulse.waveforms import (
    ConstantWaveform,
    DragGaussianWaveform,
    GaussianWaveform,
    Waveform,
)
from braket.timings.time_series import TimeSeries


@dataclass
class _FrameState:
    dt: float
    frequency: float = 0
    phase: float = 0
    current_time: float = 0
    amplitude: float = 0
    scale: float = 1


@dataclass
class _ParseState:
    variables: dict
    frame_data: dict[str, _FrameState]


class _ApproximationParser(QASMVisitor[_ParseState]):
    """Walk the AST and build the output signal amplitude, frequency and phases
    for each channel.
    """

    TIME_UNIT_TO_EXP: ClassVar = {"dt": 4, "ns": 3, "us": 2, "ms": 1, "s": 0}

    def __init__(self, program: Program, frames: dict[str, Frame]):
        self.amplitudes = defaultdict(TimeSeries)
        self.frequencies = defaultdict(TimeSeries)
        self.phases = defaultdict(TimeSeries)
        context = _ParseState(variables={}, frame_data=_init_frame_data(frames))
        self._qubit_frames_mapping: dict[str, list[str]] = _init_qubit_frame_mapping(frames)
        self.visit(program.to_ast(include_externs=False), context)

    def visit(
        self, node: Union[ast.QASMNode, ast.Expression], context: Optional[_ParseState] = None
    ) -> Any:
        """Visit a node.

        Args:
            node (Union[ast.QASMNode, ast.Expression]): The node to visit.
            context (Optional[_ParseState]): The parse state context.

        Returns:
            Any: The parsed return value.
        """
        return super().visit(node, context)

    def _get_frame_parameters(
        self, parameters: list[ast.Expression], context: _ParseState
    ) -> Union[KeysView, list[str]]:
        frame_ids = set()
        for expression in parameters:
            identifier_name = self.visit(expression, context)
            if match := re.search(r"^\$[0-9]+$", identifier_name):
                qubit_number = match.group()[1:]
                frame_ids.update(self._qubit_frames_mapping.get(qubit_number, []))
            else:
                frame_ids.add(identifier_name)
        return frame_ids

    def _delay_frame(self, frame_id: str, to_delay_time: float, context: _ParseState) -> None:
        frame_data = context.frame_data[frame_id]
        if to_delay_time >= frame_data.current_time + frame_data.dt:
            start_time = frame_data.current_time
            self.amplitudes[frame_id].put(start_time, 0)
            self.frequencies[frame_id].put(start_time, frame_data.frequency)
            self.phases[frame_id].put(start_time, frame_data.phase)
        if to_delay_time >= frame_data.current_time + (2 * frame_data.dt):
            end_time = to_delay_time - frame_data.dt
            self.amplitudes[frame_id].put(end_time, 0)
            self.frequencies[frame_id].put(end_time, frame_data.frequency)
            self.phases[frame_id].put(end_time, frame_data.phase)
        context.frame_data[frame_id].current_time = to_delay_time

    def visit_Program(self, node: ast.Program, context: _ParseState = None) -> None:
        """Visit a Program.

        Args:
            node (ast.Program): The program.
            context (_ParseState): The parse state context.
        """
        for statement in node.statements:
            self.visit(statement, context)

    def visit_ExpressionStatement(self, node: ast.ExpressionStatement, context: _ParseState) -> Any:
        """Visit an Expression.

        Args:
            node (ast.ExpressionStatement): The expression.
            context (_ParseState): The parse state context.

        Returns:
            Any: The parsed return value.
        """
        return self.visit(node.expression, context)  # need to check

    def visit_ClassicalDeclaration(
        self, node: ast.ClassicalDeclaration, context: _ParseState
    ) -> Union[dict, None]:
        """Visit a Classical Declaration.
            node.type, node.identifier, node.init_expression
            angle[20] a = 1+2;
            waveform wf = [];
            port a;

        Args:
            node (ast.ClassicalDeclaration): The classical declaration.
            context (_ParseState): The parse state context.

        Raises:
            NotImplementedError: Raised if the node is not a PortType, FrameType, or
                WaveformType.

        Returns:
            Union[dict, None]: Returns a dict if WaveformType, None otherwise.
        """
        identifier = self.visit(node.identifier, context)
        if type(node.type) == ast.WaveformType:
            context.variables[identifier] = self.visit(node.init_expression, context)
        elif type(node.type) == ast.FrameType:
            pass
        elif type(node.type) == ast.PortType:
            pass
        else:
            raise NotImplementedError

    def visit_DelayInstruction(self, node: ast.DelayInstruction, context: _ParseState) -> None:
        """Visit a Delay Instruction.
            node.duration, node.qubits
            delay[100ns] $0;

        Args:
            node (ast.DelayInstruction): The classical declaration.
            context (_ParseState): The parse state context.
        """
        duration = self.visit(node.duration, context)
        frames = self._get_frame_parameters(node.qubits, context)
        if len(frames) == 0:
            # barrier without arguments is applied to all the frames of the context
            frames = list(context.frame_data.keys())
        dts = [context.frame_data[frame_id].dt for frame_id in frames]
        max_time = max([context.frame_data[frame_id].current_time for frame_id in frames])
        # All frames are delayed till the first multiple of the LCM([port.dts])
        # after the longest time of all considered frames
        lcm = _lcm_floats(*dts)
        barrier_time = _ceil_approx(max_time / lcm) * lcm

        for frame_id in frames:
            self._delay_frame(frame_id, barrier_time + duration, context)

    def visit_QuantumBarrier(self, node: ast.QuantumBarrier, context: _ParseState) -> None:
        """Visit a Quantum Barrier.
            barrier $0;
            barrier;
            barrier frame, frame1;

        Args:
            node (ast.QuantumBarrier): The quantum barrier.
            context (_ParseState): The parse state context.

        Returns:
            None: No return value.
        """
        frames = self._get_frame_parameters(node.qubits, context)
        if len(frames) == 0:
            # barrier without arguments is applied to all the frames of the context
            frames = list(context.frame_data.keys())
        dts = [context.frame_data[frame_id].dt for frame_id in frames]
        max_time = max([context.frame_data[frame_id].current_time for frame_id in frames])
        # All frames are delayed till the first multiple of the LCM([port.dts])
        # after the longest time of all considered frames
        lcm = _lcm_floats(*dts)
        barrier_time = _ceil_approx(max_time / lcm) * lcm

        for frame_id in frames:
            self._delay_frame(frame_id, barrier_time, context)

    def visit_FunctionCall(self, node: ast.FunctionCall, context: _ParseState) -> Any:
        """Visit a Quantum Barrier.
            node.name, node.arguments
            f(args,arg2)

        Args:
            node (ast.FunctionCall): The function call.
            context (_ParseState): The parse state context.

        Returns:
            Any: The parsed return value.
        """
        func_name = node.name.name
        return getattr(self, func_name)(node, context)

    def visit_Identifier(self, node: ast.Identifier, context: _ParseState) -> Any:
        """Visit Identifier.
            node.name
            x
        Args:
            node (ast.Identifier): The identifier.
            context (_ParseState): The parse state context.

        Returns:
            Any: The parsed return value.
        """
        if node.name in context.variables:
            return context.variables[node.name]
        else:
            return node.name

    def visit_UnaryExpression(self, node: ast.UnaryExpression, context: _ParseState) -> bool:
        """Visit Unary Expression.
            node.op, node.expression
            ~ ! -

        Args:
            node (ast.UnaryExpression): The unary expression.
            context (_ParseState): The parse state context.

        Returns:
            bool: The parsed boolean operator.

        Raises:
            NotImplementedError: Raised for unsupported boolean operators.
        """
        if node.op == ast.UnaryOperator["-"]:
            return -1 * self.visit(node.expression, context)
        elif node.op == ast.UnaryOperator["!"]:
            return not self.visit(node.expression, context)
        elif node.op == ast.UnaryOperator["~"]:
            return ~self.visit(node.expression, context)
        else:
            raise NotImplementedError

    # flake8: noqa: C901
    def visit_BinaryExpression(self, node: ast.BinaryExpression, context: _ParseState) -> Any:
        """Visit Binary Expression.
            node.lhs, node.rhs, node.op
            1+2
            a.b
            > < >= <= == != && || | ^ & << >> + - * / % ** .

        Args:
            node (ast.BinaryExpression): The binary expression.
            context (_ParseState): The parse state context.

        Raises:
            NotImplementedError: Raised if the binary operator is not in
                [> < >= <= == != && || | ^ & << >> + - * / % ** ]

        Returns:
            Any: The parsed binary operator.
        """
        lhs = self.visit(node.lhs, context)
        rhs = self.visit(node.rhs, context)

        op = ast.BinaryOperator

        if node.op == op["+"]:
            return lhs + rhs
        elif node.op == op["-"]:
            return lhs - rhs
        elif node.op == op["*"]:
            return lhs * rhs
        elif node.op == op["/"]:
            return lhs / rhs
        elif node.op == op["%"]:
            return lhs % rhs
        elif node.op == op["**"]:
            return lhs**rhs
        elif node.op == op[">"]:
            return lhs > rhs
        elif node.op == op["<"]:
            return lhs < rhs
        elif node.op == op[">="]:
            return lhs >= rhs
        elif node.op == op["<="]:
            return lhs <= rhs
        elif node.op == op["=="]:
            return lhs == rhs
        elif node.op == op["!="]:
            return lhs != rhs
        elif node.op == op["&&"]:
            return lhs and rhs
        elif node.op == op["||"]:
            return lhs or rhs
        elif node.op == op["|"]:
            return lhs | rhs
        elif node.op == op["^"]:
            return lhs ^ rhs
        elif node.op == op["&"]:
            return lhs & rhs
        elif node.op == op["<<"]:
            return lhs << rhs
        elif node.op == op[">>"]:
            return lhs >> rhs
        else:
            raise NotImplementedError

    def visit_ArrayLiteral(self, node: ast.ArrayLiteral, context: _ParseState) -> list[Any]:
        """Visit Array Literal.
            node.values
            {1,2,4}

        Args:
            node (ast.ArrayLiteral): The array literal.
            context (_ParseState): The parse state context.

        Returns:
            list[Any]: The parsed ArrayLiteral.
        """
        return [self.visit(e, context) for e in node.values]

    def visit_IntegerLiteral(self, node: ast.IntegerLiteral, context: _ParseState) -> int:
        """Visit Integer Literal.
            node.value
            1
        Args:
            node (ast.IntegerLiteral): The integer literal.
            context (_ParseState): The parse state context.

        Returns:
            int: The parsed int value.
        """
        return int(node.value)

    def visit_ImaginaryLiteral(self, node: ast.ImaginaryLiteral, context: _ParseState) -> complex:
        """Visit Imaginary Number Literal.
            node.value
            1.3im
        Args:
            node (ast.ImaginaryLiteral): The imaginary number literal.
            context (_ParseState): The parse state context.

        Returns:
            complex: The parsed complex value.
        """
        return complex(node.value * 1j)

    def visit_FloatLiteral(self, node: ast.FloatLiteral, context: _ParseState) -> float:
        """Visit Float Literal.
            node.value
            1.1
        Args:
            node (ast.FloatLiteral): The float literal.
            context (_ParseState): The parse state context.

        Returns:
            float: The parsed float value.
        """
        return float(node.value)

    def visit_BooleanLiteral(self, node: ast.BooleanLiteral, context: _ParseState) -> bool:
        """Visit Boolean Literal.
            node.value
            true
        Args:
            node (ast.BooleanLiteral): The boolean literal.
            context (_ParseState): The parse state context.

        Returns:
            bool: The parsed boolean value.
        """
        return True if node.value else False

    def visit_DurationLiteral(self, node: ast.DurationLiteral, context: _ParseState) -> float:
        """Visit Duration Literal.
            node.value, node.unit (node.unit.name, node.unit.value)
            1
        Args:
            node (ast.DurationLiteral): The duration literal.
            context (_ParseState): The parse state context.

        Raises:
            ValueError: Raised based on time unit not being in `self.TIME_UNIT_TO_EXP`.

        Returns:
            float: The duration represented as a float
        """
        if node.unit.name not in self.TIME_UNIT_TO_EXP:
            raise ValueError(f"Unexpected duration specified: {node.unit.name}:{node.unit.value}")
        multiplier = 10 ** (-3 * self.TIME_UNIT_TO_EXP[node.unit.name])
        return multiplier * node.value

    # The following are function call declarations supported by the parser.

    def set_frequency(self, node: ast.FunctionCall, context: _ParseState) -> None:
        """A 'set_frequency' Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.
        """
        frame = self.visit(node.arguments[0], context)
        value = self.visit(node.arguments[1], context)
        context.frame_data[frame].frequency = value

    def shift_frequency(self, node: ast.FunctionCall, context: _ParseState) -> None:
        """A 'shift_frequency' Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.
        """
        frame = self.visit(node.arguments[0], context)
        value = self.visit(node.arguments[1], context)
        context.frame_data[frame].frequency += value

    def set_phase(self, node: ast.FunctionCall, context: _ParseState) -> None:
        """A 'set_phase' Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.
        """
        frame = self.visit(node.arguments[0], context)
        value = self.visit(node.arguments[1], context)
        context.frame_data[frame].phase = value % (2 * np.pi)

    def shift_phase(self, node: ast.FunctionCall, context: _ParseState) -> None:
        """A 'shift_phase' Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.
        """
        frame = self.visit(node.arguments[0], context)
        value = self.visit(node.arguments[1], context)
        context.frame_data[frame].phase += value
        context.frame_data[frame].phase %= 2 * np.pi

    def set_scale(self, node: ast.FunctionCall, context: _ParseState) -> None:
        """A 'set_scale' Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.
        """
        frame = self.visit(node.arguments[0], context)
        value = self.visit(node.arguments[1], context)
        context.frame_data[frame].scale = value

    def capture_v0(self, node: ast.FunctionCall, context: _ParseState) -> None:
        """A 'capture_v0' Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.
        """
        pass

    def play(self, node: ast.FunctionCall, context: _ParseState) -> None:
        """A 'play' Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.

        Raises:
            NotImplementedError: Raises if not of type
                [ast.Identifier, ast.FunctionCall, ast.ArrayLiteral]

        Returns:
            None: Returns None
        """
        frame_id = self.visit(node.arguments[0], context)
        if isinstance(node.arguments[1], ast.ArrayLiteral):
            amps = self.visit(node.arguments[1], context)
        elif isinstance(node.arguments[1], (ast.Identifier, ast.FunctionCall)):
            amps = self.visit(node.arguments[1], context)
            if isinstance(amps, Waveform):
                amps = amps.sample(context.frame_data[frame_id].dt)
            elif isinstance(amps, str):
                raise NameError(f"waveform '{amps}' is not defined.")
        else:
            raise NotImplementedError
        frame_data = context.frame_data[frame_id]
        for value in amps:
            self.amplitudes[frame_id].put(
                frame_data.current_time, complex(frame_data.scale * value)
            )
            self.frequencies[frame_id].put(frame_data.current_time, frame_data.frequency)
            self.phases[frame_id].put(frame_data.current_time, frame_data.phase)
            frame_data.current_time += frame_data.dt

    def constant(self, node: ast.FunctionCall, context: _ParseState) -> Waveform:
        """A 'constant' Waveform Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.

        Returns:
            Waveform: The waveform object representing the function call.
        """
        args = [self.visit(arg, context) for arg in node.arguments]
        return ConstantWaveform(*args)

    def gaussian(self, node: ast.FunctionCall, context: _ParseState) -> Waveform:
        """A 'gaussian' Waveform Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.

        Returns:
            Waveform: The waveform object representing the function call.
        """
        args = [self.visit(arg, context) for arg in node.arguments]
        return GaussianWaveform(*args)

    def drag_gaussian(self, node: ast.FunctionCall, context: _ParseState) -> Waveform:
        """A 'drag_gaussian' Waveform Function call.

        Args:
            node (ast.FunctionCall): The function call node.
            context (_ParseState): The parse state.

        Returns:
            Waveform: The waveform object representing the function call.
        """
        args = [self.visit(arg, context) for arg in node.arguments]
        return DragGaussianWaveform(*args)


def _init_frame_data(frames: dict[str, Frame]) -> dict[str, _FrameState]:
    frame_states = {}
    for frameId, frame in frames.items():
        frame_states[frameId] = _FrameState(
            frame.port.dt, frame.frequency, frame.phase % (2 * np.pi)
        )
    return frame_states


def _init_qubit_frame_mapping(frames: dict[str, Frame]) -> dict[str, list[str]]:
    mapping = {}
    for frameId in frames.keys():
        if m := (
            re.search(r"q(\d+)_q(\d+)_[a-z_]+", frameId) or re.search(r"[rq](\d+)_[a-z_]+", frameId)
        ):
            for qubit in m.groups():
                if qubit in mapping:
                    mapping[qubit].append(frameId)
                else:
                    mapping[qubit] = [frameId]
    return mapping


def _lcm_floats(*dts: list[float]) -> float:
    """Return the least common multiple of time increments of a list of frames
        A time increment is the inverse of the corresponding sample rate which is considered
        an integer.
        LCM of rational numbers is lcm = (LCM of numerators) / (GCD of denominators)
        Hence the LCM of dts is 1/gcd([sample rates])

    Args:
        *dts (list[float]): list of time resolutions

    Returns:
        float: The LCM of time increments for a list of frames.
    """
    sample_rates = [round(1 / dt) for dt in dts]
    res_gcd = sample_rates[0]
    for sr in sample_rates[1:]:
        res_gcd = np.gcd(res_gcd, sr)
    return 1 / res_gcd


def _ceil_approx(number: float) -> int:
    return int(number) + 1 if abs(number - int(number)) > 0.001 else int(number)



import math

import numpy as np
import pytest
from gate_model_device_testing_utils import get_tol

from braket.aws import AwsDevice
from braket.circuits import Circuit, Noise, Observable

SHOTS = 500
DM1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/dm1"
SIMULATOR_ARNS = [DM1_ARN]


@pytest.mark.parametrize("simulator_arn", SIMULATOR_ARNS)
def test_mixed_states(simulator_arn, aws_session, s3_destination_folder):
    num_qubits = 10
    circuit = _mixed_states(num_qubits)
    device = AwsDevice(simulator_arn, aws_session)

    tol = get_tol(SHOTS)
    result = device.run(circuit, shots=SHOTS, s3_destination_folder=s3_destination_folder).result()
    probabilities = result.measurement_probabilities
    probability_sum = 0
    for bitstring in probabilities:
        assert probabilities[bitstring] >= 0
        probability_sum += probabilities[bitstring]
    assert math.isclose(probability_sum, 1, rel_tol=tol["rtol"], abs_tol=tol["atol"])
    assert len(result.measurements) == SHOTS


def _mixed_states(n_qubits: int) -> Circuit:
    noise = Noise.PhaseFlip(probability=0.2)
    circ = Circuit()
    for qubit in range(0, n_qubits - 2, 3):
        circ.x(qubit).y(qubit + 1).cnot(qubit, qubit + 2).x(qubit + 1).z(qubit + 2)
        circ.apply_gate_noise(noise, target_qubits=[qubit, qubit + 2])

    # attach the result types
    circ.probability()
    circ.expectation(observable=Observable.Z(), target=0)

    return circ


@pytest.mark.parametrize("simulator_arn", SIMULATOR_ARNS)
def test_density_matrix_result_type(simulator_arn):
    circ = Circuit().bit_flip(0, 0.2).density_matrix()
    device = AwsDevice(simulator_arn)
    density_matrix_result = device.run(circ).result().result_types[0].value
    assert np.allclose(
        density_matrix_result,
        [[[0.8, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.2, 0.0]]],
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class QueueType(str, Enum):
    """Enumerates the possible priorities for the queue.

    Values:
        NORMAL: Represents normal queue for the device.
        PRIORITY: Represents priority queue for the device.
    """

    NORMAL = "Normal"
    PRIORITY = "Priority"


@dataclass()
class QueueDepthInfo:
    """Represents quantum tasks and hybrid jobs queue depth information.

    Attributes:
        quantum_tasks (dict[QueueType, str]): number of quantum tasks waiting
            to run on a device. This includes both 'Normal' and 'Priority' tasks.
            For Example, {'quantum_tasks': {QueueType.NORMAL: '7', QueueType.PRIORITY: '3'}}
        jobs (str): number of hybrid jobs waiting to run on a device. Additionally, for QPUs if
            hybrid jobs queue depth is 0, we display information about priority and count of the
            running hybrid jobs. Example, 'jobs': '0 (1 prioritized job(s) running)'
    """

    quantum_tasks: dict[QueueType, str]
    jobs: str


@dataclass
class QuantumTaskQueueInfo:
    """Represents quantum tasks queue information.

    Attributes:
        queue_type (QueueType): type of the quantum_task queue either 'Normal'
            or 'Priority'.
        queue_position (Optional[str]): current position of your quantum task within a respective
            device queue. This value can be None based on the state of the task. Default: None.
        message (Optional[str]): Additional message information. This key is present only
            if 'queue_position' is None. Default: None.
    """

    queue_type: QueueType
    queue_position: Optional[str] = None
    message: Optional[str] = None


@dataclass
class HybridJobQueueInfo:
    """Represents hybrid job queue information.

    Attributes:
        queue_position (Optional[str]): current position of your hybrid job within a respective
            device queue. If the queue position of the hybrid job is greater than 15, we
            return '>15' as the queue_position return value. The queue_position is only
            returned when hybrid job is not in RUNNING/CANCELLING/TERMINAL states, else
            queue_position is returned as None.
        message (Optional[str]): Additional message information. This key is present only
            if 'queue_position' is None. Default: None.
    """

    queue_position: Optional[str] = None
    message: Optional[str] = None



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re

import pytest
from botocore.exceptions import ClientError

from braket.aws.aws_device import AwsDevice
from braket.circuits.circuit import Circuit
from braket.devices import LocalSimulator

DEVICE = LocalSimulator()
SHOTS = 8000

IONQ_ARN = "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony"
SIMULATOR_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
OQC_ARN = "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy"


@pytest.mark.parametrize("arn", [(IONQ_ARN), (SIMULATOR_ARN)])
def test_unsupported_devices(arn):
    device = AwsDevice(arn)
    if device.status == "OFFLINE":
        pytest.skip("Device offline")

    circ = Circuit().h(0).cnot(0, 1).h(2).measure([0, 1])
    error_string = re.escape(
        "An error occurred (ValidationException) when calling the "
        "CreateQuantumTask operation: Device requires all qubits in the program to be measured. "
        "This may be caused by declaring non-contiguous qubits or measuring partial qubits"
    )
    with pytest.raises(ClientError, match=error_string):
        device.run(circ, shots=1000)


@pytest.mark.parametrize("sim", [("braket_sv"), ("braket_dm")])
def test_measure_on_local_sim(sim):
    circ = Circuit().h(0).cnot(0, 1).h(2).measure([0, 1])
    device = LocalSimulator(sim)
    result = device.run(circ, SHOTS).result()
    assert len(result.measurements[0]) == 2
    assert result.measured_qubits == [0, 1]


@pytest.mark.parametrize("arn", [(OQC_ARN)])
def test_measure_on_supported_devices(arn):
    device = AwsDevice(arn)
    if not device.is_available:
        pytest.skip("Device offline")
    circ = Circuit().h(0).cnot(0, 1).measure([0])
    result = device.run(circ, SHOTS).result()
    assert len(result.measurements[0]) == 1
    assert result.measured_qubits == [0]


@pytest.mark.parametrize(
    "circuit, expected_measured_qubits",
    [
        (Circuit().h(0).cnot(0, 1).cnot(1, 2).cnot(2, 3).measure([0, 1, 3]), [0, 1, 3]),
        (Circuit().h(0).measure(0), [0]),
    ],
)
def test_measure_targets(circuit, expected_measured_qubits):
    result = DEVICE.run(circuit, SHOTS).result()
    assert result.measured_qubits == expected_measured_qubits
    assert len(result.measurements[0]) == len(expected_measured_qubits)


def test_measure_with_noise():
    device = LocalSimulator("braket_dm")
    circuit = Circuit().x(0).x(1).bit_flip(0, probability=0.1).measure(0)
    result = device.run(circuit, SHOTS).result()
    assert result.measured_qubits == [0]
    assert len(result.measurements[0]) == 1



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
from unittest.mock import Mock

from braket.aws import AwsQuantumTaskBatch

DWAVE_ARN = "arn:aws:braket:::device/qpu/d-wave/Advantage_system1"
RIGETTI_ARN = "arn:aws:braket:::device/qpu/rigetti/Aspen-10"
IONQ_ARN = "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony"
OQC_ARN = "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy"
SV1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
DM1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/dm1"
TN1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/tn1"
XANADU_ARN = "arn:aws:braket:us-east-1::device/qpu/xanadu/Borealis"

RIGETTI_REGION = "us-west-1"


class MockS3:
    MOCK_TASK_METADATA = {
        "braketSchemaHeader": {"name": "braket.task_result.task_metadata", "version": "1"},
        "id": "task_arn",
        "shots": 100,
        "deviceId": "default",
    }

    MOCK_S3_RESULT_GATE_MODEL = json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.task_result.gate_model_task_result",
                "version": "1",
            },
            "measurements": [[0, 0], [0, 0], [0, 0], [1, 1]],
            "measuredQubits": [0, 1],
            "taskMetadata": MOCK_TASK_METADATA,
            "additionalMetadata": {
                "action": {
                    "braketSchemaHeader": {"name": "braket.ir.jaqcd.program", "version": "1"},
                    "instructions": [{"control": 0, "target": 1, "type": "cnot"}],
                },
            },
        }
    )

    MOCK_S3_RESULT_GATE_MODEL_WITH_RESULT_TYPES = json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.task_result.gate_model_task_result",
                "version": "1",
            },
            "measurements": [[0, 0], [0, 0], [0, 0], [1, 1]],
            "measuredQubits": [0, 1],
            "resultTypes": [
                {
                    "type": {"observable": ["h", "x"], "targets": [0, 1], "type": "expectation"},
                    "value": 0.7071067811865474,
                },
                {
                    "type": {"states": ["01", "10", "00", "11"], "type": "amplitude"},
                    "value": {
                        "01": [0.0, 0.0],
                        "10": [0.0, 0.0],
                        "00": [0.7071067811865475, 0.0],
                        "11": [0.7071067811865475, 0.0],
                    },
                },
            ],
            "taskMetadata": MOCK_TASK_METADATA,
            "additionalMetadata": {
                "action": {
                    "braketSchemaHeader": {"name": "braket.ir.jaqcd.program", "version": "1"},
                    "instructions": [{"control": 0, "target": 1, "type": "cnot"}],
                },
            },
        }
    )

    MOCK_S3_RESULT_ANNEALING = json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.task_result.annealing_task_result",
                "version": "1",
            },
            "solutions": [[-1, -1, -1, -1], [1, -1, 1, 1], [1, -1, -1, 1]],
            "solutionCounts": [3, 2, 4],
            "values": [0.0, 1.0, 2.0],
            "variableCount": 4,
            "taskMetadata": {
                "id": "task_arn",
                "shots": 100,
                "deviceId": DWAVE_ARN,
            },
            "additionalMetadata": {
                "action": {
                    "type": "ISING",
                    "linear": {"0": 0.3333, "1": -0.333, "4": -0.333, "5": 0.333},
                    "quadratic": {"0,4": 0.667, "0,5": -1.0, "1,4": 0.667, "1,5": 0.667},
                },
                "dwaveMetadata": {
                    "activeVariables": [0],
                    "timing": {
                        "qpuSamplingTime": 100,
                        "qpuAnnealTimePerSample": 20,
                        "qpuAccessTime": 10917,
                        "qpuAccessOverheadTime": 3382,
                        "qpuReadoutTimePerSample": 274,
                        "qpuProgrammingTime": 9342,
                        "qpuDelayTimePerSample": 21,
                        "postProcessingOverheadTime": 117,
                        "totalPostProcessingTime": 117,
                        "totalRealTime": 10917,
                        "runTimeChip": 1575,
                        "annealTimePerRun": 20,
                        "readoutTimePerRun": 274,
                    },
                },
            },
        }
    )

    MOCK_S3_RESULT_PHOTONIC_MODEL = json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.task_result.photonic_model_task_result",
                "version": "1",
            },
            "measurements": [[[1, 2, 3, 4]], [[4, 3, 2, 1]], [[0, 0, 0, 0]]],
            "taskMetadata": {
                "id": "task_arn",
                "shots": 3,
                "deviceId": XANADU_ARN,
            },
            "additionalMetadata": {
                "action": {
                    "source": "Vac | q[0]",
                },
                "xanaduMetadata": {
                    "compiledProgram": "DECLARE ro BIT[2];",
                },
            },
        }
    )

    MOCK_S3_RESULT_ANALOG_HAMILTONIAN_SIMULTION = json.dumps(
        {
            "braketSchemaHeader": {
                "name": "braket.task_result.analog_hamiltonian_simulation_task_result",
                "version": "1",
            },
            "taskMetadata": {
                "id": "task_arn",
                "shots": 3,
                "deviceId": "mock_arn",
            },
            "measurements": [
                {
                    "shotMetadata": {"shotStatus": "Success"},
                    "shotResult": {
                        "preSequence": [1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
                        "postSequence": [0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0],
                    },
                },
                {
                    "shotMetadata": {"shotStatus": "Partial Success"},
                    "shotResult": {
                        "preSequence": [1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
                        "postSequence": None,
                    },
                },
                {
                    "shotMetadata": {"shotStatus": "Failure"},
                    "shotResult": {"preSequence": None, "postSequence": None},
                },
            ],
        }
    )


def run_and_assert(
    aws_quantum_task_mock,
    device,
    default_s3_folder,
    default_shots,
    default_poll_timeout,
    default_poll_interval,
    circuit,
    s3_destination_folder,  # Treated as positional arg
    shots,  # Treated as positional arg
    poll_timeout_seconds,  # Treated as positional arg
    poll_interval_seconds,  # Treated as positional arg
    inputs,  # Treated as positional arg
    gate_definitions,  # Treated as positional arg
    reservation_arn,  # Treated as positional arg
    extra_args,
    extra_kwargs,
):
    task_mock = Mock()
    aws_quantum_task_mock.return_value = task_mock

    run_args = []
    if s3_destination_folder is not None:
        run_args.append(s3_destination_folder)
    if shots is not None:
        run_args.append(shots)
    if poll_timeout_seconds is not None:
        run_args.append(poll_timeout_seconds)
    if poll_interval_seconds is not None:
        run_args.append(poll_interval_seconds)
    if inputs is not None:
        run_args.append(inputs)
    if gate_definitions is not None:
        run_args.append(gate_definitions)
    run_args += extra_args if extra_args else []
    run_kwargs = extra_kwargs or {}
    if reservation_arn:
        run_kwargs.update({"reservation_arn": reservation_arn})

    task = device.run(circuit, *run_args, **run_kwargs)
    assert task == task_mock

    create_args, create_kwargs = _create_task_args_and_kwargs(
        default_s3_folder,
        default_shots,
        default_poll_timeout,
        default_poll_interval,
        s3_destination_folder,
        shots,
        poll_timeout_seconds,
        poll_interval_seconds,
        inputs,
        gate_definitions,
        reservation_arn,
        extra_args,
        extra_kwargs,
    )

    aws_quantum_task_mock.assert_called_with(
        device._aws_session, device.arn, circuit, *create_args, **create_kwargs
    )


def run_batch_and_assert(
    aws_quantum_task_mock,
    aws_session_mock,
    device,
    default_s3_folder,
    default_shots,
    default_poll_timeout,
    default_poll_interval,
    circuits,
    s3_destination_folder,
    shots,
    max_parallel,
    max_connections,
    poll_timeout_seconds,
    poll_interval_seconds,
    inputs,
    gate_definitions,
    reservation_arn,
    extra_args,
    extra_kwargs,
):
    task_mock = Mock()
    task_mock.state.return_value = "COMPLETED"
    aws_quantum_task_mock.return_value = task_mock
    new_session_mock = Mock()
    aws_session_mock.return_value = new_session_mock

    run_args = []
    if s3_destination_folder is not None:
        run_args.append(s3_destination_folder)
    if shots is not None:
        run_args.append(shots)
    if max_parallel is not None:
        run_args.append(max_parallel)
    if max_connections is not None:
        run_args.append(max_connections)
    if poll_timeout_seconds is not None:
        run_args.append(poll_timeout_seconds)
    if poll_interval_seconds is not None:
        run_args.append(poll_interval_seconds)
    if inputs is not None:
        run_args.append(inputs)
    if gate_definitions is not None:
        run_args.append(gate_definitions)
    run_args += extra_args if extra_args else []
    run_kwargs = extra_kwargs or {}
    if reservation_arn:
        run_kwargs.update({"reservation_arn": reservation_arn})

    batch = device.run_batch(circuits, *run_args, **run_kwargs)
    assert batch.tasks == [task_mock for _ in range(len(circuits))]

    create_args, create_kwargs = _create_task_args_and_kwargs(
        default_s3_folder,
        default_shots,
        default_poll_timeout,
        default_poll_interval,
        s3_destination_folder,
        shots,
        poll_timeout_seconds,
        poll_interval_seconds,
        inputs,
        gate_definitions,
        reservation_arn,
        extra_args,
        extra_kwargs,
    )

    max_pool_connections = max_connections or AwsQuantumTaskBatch.MAX_CONNECTIONS_DEFAULT

    # aws_session_mock.call_args.kwargs syntax is newer than Python 3.7
    assert aws_session_mock.call_args[1]["config"].max_pool_connections == max_pool_connections
    aws_quantum_task_mock.assert_any_call(
        new_session_mock, device.arn, circuits[0], *create_args, **create_kwargs
    )
    aws_quantum_task_mock.assert_any_call(
        new_session_mock, device.arn, circuits[1], *create_args, **create_kwargs
    )


def _create_task_args_and_kwargs(
    default_s3_folder,
    default_shots,
    default_poll_timeout,
    default_poll_interval,
    s3_folder,
    shots,
    poll_timeout_seconds,
    poll_interval_seconds,
    inputs,
    gate_definitions,
    reservation_arn,
    extra_args,
    extra_kwargs,
):
    create_args = [
        s3_folder if s3_folder is not None else default_s3_folder,
        shots if shots is not None else default_shots,
    ]
    create_args += extra_args if extra_args else []
    create_kwargs = extra_kwargs or {}
    create_kwargs.update(
        {
            "poll_timeout_seconds": (
                poll_timeout_seconds if poll_timeout_seconds is not None else default_poll_timeout
            ),
            "poll_interval_seconds": (
                poll_interval_seconds
                if poll_interval_seconds is not None
                else default_poll_interval
            ),
            "inputs": inputs,
            "gate_definitions": gate_definitions,
            "reservation_arn": reservation_arn,
        }
    )
    return create_args, create_kwargs



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import importlib
import sys
from unittest.mock import patch

import braket.tasks


@patch("braket.ipython_utils.running_in_jupyter")
def test_nest_asyncio_not_applied(running_in_jupyter):
    running_in_jupyter.return_value = False
    importlib.reload(braket.tasks)
    assert "nest_asyncio" not in sys.modules


@patch("braket.ipython_utils.running_in_jupyter")
def test_nest_asyncio_is_applied(running_in_jupyter):
    running_in_jupyter.return_value = True
    importlib.reload(braket.tasks)
    assert "nest_asyncio" in sys.modules



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections.abc import Iterable
from typing import Any, Optional, Union

from braket.circuits.noise_model.criteria import Criteria, CriteriaKey, CriteriaKeyResult
from braket.circuits.noise_model.criteria_input_parsing import (
    parse_operator_input,
    parse_qubit_input,
)
from braket.circuits.noise_model.result_type_criteria import ResultTypeCriteria
from braket.circuits.observable import Observable
from braket.circuits.result_type import ObservableResultType, ResultType
from braket.registers.qubit_set import QubitSetInput


class ObservableCriteria(ResultTypeCriteria):
    """This class models noise Criteria based on the Braket SDK Observable classes."""

    def __init__(
        self,
        observables: Optional[Union[Observable, Iterable[Observable]]] = None,
        qubits: Optional[QubitSetInput] = None,
    ):
        """Creates Observable-based Criteria. See instruction_matches() for more details.

        Args:
            observables (Optional[Union[Observable, Iterable[Observable]]]): A set of relevant
                Observables. Observables must only operate on a single qubit. Optional. If
                observables are not specified, this criteria will match on any observable.
            qubits (Optional[QubitSetInput]): A set of relevant qubits. If no qubits
                are provided, all (possible) qubits are considered to be relevant.

        Throws:
            ValueError: If the operators operate on more than one qubit.
        """
        self._observables = parse_operator_input(observables)
        self._qubits = parse_qubit_input(qubits, 1)

    def __str__(self):
        observables_names = (
            {observable.__name__ for observable in self._observables}
            if self._observables is not None
            else None
        )
        return f"{self.__class__.__name__}({observables_names}, {self._qubits})"

    def __repr__(self):
        observables_names = (
            {observable.__name__ for observable in self._observables}
            if self._observables is not None
            else None
        )
        return f"{self.__class__.__name__}(observables={observables_names}, qubits={self._qubits})"

    def applicable_key_types(self) -> Iterable[CriteriaKey]:
        """Returns an Iterable of criteria keys.

        Returns:
            Iterable[CriteriaKey]: This Criteria operates on Observables and Qubits.
        """
        return [CriteriaKey.OBSERVABLE, CriteriaKey.QUBIT]

    def get_keys(self, key_type: CriteriaKey) -> Union[CriteriaKeyResult, set[Any]]:
        """Gets the keys for a given CriteriaKey.

        Args:
            key_type (CriteriaKey): The relevant Criteria Key.

        Returns:
            Union[CriteriaKeyResult, set[Any]]: The return value is based on the key type:
            OBSERVABLE will return a set of Observable classes that are relevant to this Criteria,
            or CriteriaKeyResult.ALL if the Criteria is relevant for all (possible) observables.
            QUBIT will return a set of qubit targets that are relevant to this Criteria, or
            CriteriaKeyResult.ALL if the Criteria is relevant for all (possible) qubits.
            All other keys will return an empty set.
        """
        if key_type == CriteriaKey.OBSERVABLE:
            return CriteriaKeyResult.ALL if self._observables is None else set(self._observables)
        if key_type == CriteriaKey.QUBIT:
            return CriteriaKeyResult.ALL if self._qubits is None else set(self._qubits)
        return set()

    def to_dict(self) -> dict:
        """Converts a dictionary representing an object of this class into an instance of
        this class.

        Returns:
            dict: A dictionary representing the serialized version of this Criteria.
        """
        observables = (
            [observable.__name__ for observable in self._observables]
            if self._observables is not None
            else None
        )
        qubits = list(self._qubits) if self._qubits is not None else None
        return {
            "__class__": self.__class__.__name__,
            "observables": observables,
            "qubits": qubits,
        }

    def result_type_matches(self, result_type: ResultType) -> bool:
        """Returns true if a result type matches the criteria.

        Args:
            result_type (ResultType): A result type or list of result types to match.

        Returns:
            bool: Returns true if the result type is one of the Observables provided in the
            constructor and the target is a qubit (or set of qubits)provided in the constructor.
            If observables were not provided in the constructor, then this method will accept any
            Observable.
            If qubits were not provided in the constructor, then this method will accept any
            result type target.
        """
        if not isinstance(result_type, ObservableResultType):
            return False
        if self._observables is not None and type(result_type.observable) not in self._observables:
            return False
        if self._qubits is None:
            return True
        target = list(result_type.target)
        if not target:
            return True
        return target[0] in self._qubits

    @classmethod
    def from_dict(cls, criteria: dict) -> Criteria:
        """Deserializes a dictionary into a Criteria object.

        Args:
            criteria (dict): A dictionary representation of a GateCriteria.

        Returns:
            Criteria: A deserialized GateCriteria represented by the passed in
            serialized data.
        """
        observables = (
            [getattr(Observable, name) for name in criteria["observables"]]
            if criteria["observables"] is not None
            else None
        )
        return ObservableCriteria(observables, criteria["qubits"])


Criteria.register_criteria(ObservableCriteria)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import functools

import numpy as np
import pytest

from braket.circuits.quantum_operator_helpers import (
    get_pauli_eigenvalues,
    is_cptp,
    is_hermitian,
    is_square_matrix,
    is_unitary,
    verify_quantum_operator_matrix_dimensions,
)

valid_unitary_hermitian_matrix = np.array([[0, 1], [1, 0]])

valid_CPTP_matrices = [
    np.array([[0, 1], [1, 0]]) / np.sqrt(2),
    np.array([[0, 1], [1, 0]]) / np.sqrt(2),
]

invalid_dimension_matrices = [
    (np.array([[1]])),
    (np.array([1])),
    (np.array([0, 1, 2])),
    (np.array([[0, 1], [1, 2], [3, 4]])),
    (np.array([[0, 1, 2], [2, 3]], dtype=object)),
    (np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])),
]

invalid_unitary_matrices_false = [(np.array([[0, 1], [1, 1]])), (np.array([[1, 2], [3, 4]]))]

invalid_hermitian_matrices_false = [(np.array([[1, 0], [0, 1j]])), (np.array([[1, 2], [3, 4]]))]

invalid_CPTP_matrices_false = [np.array([[1, 0], [0, 1]]), np.array([[0, 1], [1, 0]])]

invalid_matrix_type_error = np.array([[0, 1], ["a", 0]])

z_matrix = np.array([[1, 0], [0, -1]])


def test_verify_quantum_operator_matrix_dimensions():
    assert verify_quantum_operator_matrix_dimensions(valid_unitary_hermitian_matrix) is None


def test_is_unitary_true():
    assert is_unitary(valid_unitary_hermitian_matrix)


def test_is_hermitian_true():
    assert is_hermitian(valid_unitary_hermitian_matrix)


def test_is_cptp_true():
    assert is_cptp(valid_CPTP_matrices)


def test_is_square_matrix():
    assert is_square_matrix(valid_unitary_hermitian_matrix)


@pytest.mark.parametrize("matrix", invalid_dimension_matrices)
def test_verify_quantum_operator_matrix_dimensions_value_error(matrix):
    with pytest.raises(ValueError):
        verify_quantum_operator_matrix_dimensions(matrix)


@pytest.mark.parametrize("matrix", invalid_unitary_matrices_false)
def test_is_unitary_false(matrix):
    assert not is_unitary(matrix)


@pytest.mark.parametrize("matrix", invalid_hermitian_matrices_false)
def test_is_hermitian_false(matrix):
    assert not is_hermitian(matrix)


def test_is_cptp_false():
    assert not is_cptp(invalid_CPTP_matrices_false)


def test_is_hermitian_exception():
    with pytest.raises(Exception):
        is_hermitian(invalid_matrix_type_error)


def test_is_unitary_exception():
    with pytest.raises(Exception):
        is_unitary(invalid_matrix_type_error)


def test_is_cptp_exception():
    with pytest.raises(Exception):
        is_cptp([invalid_matrix_type_error])


def test_get_pauli_eigenvalues_correct_eigenvalues_one_qubit():
    """Test the get_pauli_eigenvalues function for one qubit"""
    assert np.array_equal(get_pauli_eigenvalues(1), np.diag(z_matrix))


def test_get_pauli_eigenvalues_correct_eigenvalues_two_qubits():
    """Test the get_pauli_eigenvalues function for two qubits"""
    assert np.array_equal(get_pauli_eigenvalues(2), np.diag(np.kron(z_matrix, z_matrix)))


def test_get_pauli_eigenvalues_correct_eigenvalues_three_qubits():
    """Test the get_pauli_eigenvalues function for three qubits"""
    assert np.array_equal(
        get_pauli_eigenvalues(3),
        np.diag(np.kron(z_matrix, np.kron(z_matrix, z_matrix))),
    )


@pytest.mark.parametrize("depth", list(range(1, 6)))
def test_get_pauli_eigenvalues_cache_usage(depth):
    """Test that the right number of cachings have been executed after clearing the cache"""
    get_pauli_eigenvalues.cache_clear()
    get_pauli_eigenvalues(depth)
    assert functools._CacheInfo(depth - 1, depth, 128, depth) == get_pauli_eigenvalues.cache_info()


@pytest.mark.parametrize("num_qubits", [1, 2])
def test_get_pauli_eigenvalues_immutable(num_qubits):
    with pytest.raises(ValueError):
        get_pauli_eigenvalues(num_qubits)[0] = 100



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import asyncio
from abc import ABC, abstractmethod
from typing import Any, Union

from braket.tasks.annealing_quantum_task_result import AnnealingQuantumTaskResult
from braket.tasks.gate_model_quantum_task_result import GateModelQuantumTaskResult
from braket.tasks.photonic_model_quantum_task_result import PhotonicModelQuantumTaskResult


class QuantumTask(ABC):
    """An abstraction over a quantum task on a quantum device."""

    @property
    @abstractmethod
    def id(self) -> str:
        """Get the quantum task ID.

        Returns:
            str: The quantum task ID.
        """

    @abstractmethod
    def cancel(self) -> None:
        """Cancel the quantum task."""

    @abstractmethod
    def state(self) -> str:
        """Get the state of the quantum task.

        Returns:
            str: State of the quantum task.
        """

    @abstractmethod
    def result(
        self,
    ) -> Union[
        GateModelQuantumTaskResult, AnnealingQuantumTaskResult, PhotonicModelQuantumTaskResult
    ]:
        """Get the quantum task result.

        Returns:
            Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult, PhotonicModelQuantumTaskResult]: Get
            the quantum task result. Call async_result if you want the result in an
            asynchronous way.
        """  # noqa E501

    @abstractmethod
    def async_result(self) -> asyncio.Task:
        """Get the quantum task result asynchronously.

        Returns:
            asyncio.Task: Get the quantum task result asynchronously.
        """

    def metadata(self, use_cached_value: bool = False) -> dict[str, Any]:  # noqa B027
        """Get task metadata.

        Args:
            use_cached_value (bool): If True, uses the value retrieved from the previous
                request. Default is False.

        Returns:
            dict[str, Any]: The metadata regarding the quantum task. If `use_cached_value` is True,
            then the value retrieved from the most recent request is used.
        """



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Any, Union

from braket.circuits.free_parameter import FreeParameter
from braket.circuits.observable import Observable
from braket.circuits.observables import Sum
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    SerializationProperties,
)
from braket.registers.qubit import QubitInput
from braket.registers.qubit_set import QubitSet, QubitSetInput


class ResultType:
    """Class `ResultType` represents a requested result type for the circuit.
    This class is considered the result type definition containing
    the metadata that defines what a requested result type is and what it does.
    """

    def __init__(self, ascii_symbols: list[str]):
        """Initializes a `ResultType`.

        Args:
            ascii_symbols (list[str]): ASCII string symbols for the result type. This is used when
                printing a diagram of circuits.

        Raises:
            ValueError: `ascii_symbols` is `None`
        """
        if ascii_symbols is None:
            raise ValueError("ascii_symbols must not be None")

        self._ascii_symbols = ascii_symbols

    @property
    def ascii_symbols(self) -> list[str]:
        """list[str]: Returns the ascii symbols for the requested result type."""
        return self._ascii_symbols

    @property
    def name(self) -> str:
        """Returns the name of the result type

        Returns:
            str: The name of the result type as a string
        """
        return self.__class__.__name__

    def to_ir(
        self,
        ir_type: IRType = IRType.JAQCD,
        serialization_properties: SerializationProperties | None = None,
        **kwargs,
    ) -> Any:
        """Returns IR object of the result type

        Args:
            ir_type(IRType): The IRType to use for converting the result type object to its
                IR representation. Defaults to IRType.JAQCD.
            serialization_properties (SerializationProperties | None): The serialization properties
                to use while serializing the object to the IR representation. The serialization
                properties supplied must correspond to the supplied `ir_type`. Defaults to None.

        Returns:
            Any: IR object of the result type

        Raises:
            ValueError: If the supplied `ir_type` is not supported, or if the supplied serialization
                properties don't correspond to the `ir_type`.
        """
        if ir_type == IRType.JAQCD:
            return self._to_jaqcd()
        elif ir_type == IRType.OPENQASM:
            if serialization_properties and not isinstance(
                serialization_properties, OpenQASMSerializationProperties
            ):
                raise ValueError(
                    "serialization_properties must be of type OpenQASMSerializationProperties "
                    "for IRType.OPENQASM."
                )
            return self._to_openqasm(serialization_properties or OpenQASMSerializationProperties())
        else:
            raise ValueError(f"Supplied ir_type {ir_type} is not supported.")

    def _to_jaqcd(self) -> Any:
        """Returns the JAQCD representation of the result type."""
        raise NotImplementedError("to_jaqcd has not been implemented yet.")

    def _to_openqasm(self, serialization_properties: OpenQASMSerializationProperties) -> str:
        """Returns the openqasm string representation of the result type.

        Args:
            serialization_properties (OpenQASMSerializationProperties): The serialization properties
                to use while serializing the object to the IR representation.

        Raises:
            NotImplementedError: not implemented.

        Returns:
            str: Representing the openqasm representation of the result type.
        """
        raise NotImplementedError("to_openqasm has not been implemented yet.")

    def copy(
        self,
        target_mapping: dict[QubitInput, QubitInput] | None = None,
        target: QubitSetInput | None = None,
    ) -> ResultType:
        """Return a shallow copy of the result type.

        Note:
            If `target_mapping` is specified, then `self.target` is mapped to the specified
            qubits. This is useful apply an instruction to a circuit and change the target qubits.

        Args:
            target_mapping (dict[QubitInput, QubitInput] | None): A dictionary of
                qubit mappings to apply to the target. Key is the qubit in this `target` and the
                value is what the key is changed to. Default = `None`.
            target (QubitSetInput | None): Target qubits for the new instruction.

        Returns:
            ResultType: A shallow copy of the result type.

        Raises:
            TypeError: If both `target_mapping` and `target` are supplied.

        Examples:
            >>> result_type = ResultType.Probabilities(targets=[0])
            >>> new_result_type = result_type.copy()
            >>> new_result_type.targets
            QubitSet(Qubit(0))
            >>> new_result = result_type.copy(target_mapping={0: 5})
            >>> new_result_type.target
            QubitSet(Qubit(5))
            >>> new_result = result_type.copy(target=[5])
            >>> new_result_type.target
            QubitSet(Qubit(5))
        """
        copy = self.__copy__()
        if target_mapping and target is not None:
            raise TypeError("Only 'target_mapping' or 'target' can be supplied, but not both.")
        elif target is not None:
            if hasattr(copy, "target"):
                copy.target = target
        elif hasattr(copy, "target"):
            copy.target = self._target.map(target_mapping or {})
        return copy

    @classmethod
    def register_result_type(cls, result_type: type[ResultType]) -> None:
        """Register a result type implementation by adding it into the `ResultType` class.

        Args:
            result_type (type[ResultType]): `ResultType` class to register.
        """
        setattr(cls, result_type.__name__, result_type)

    def __repr__(self) -> str:
        return f"{self.name}()"

    def __hash__(self) -> int:
        return hash(repr(self))


class ObservableResultType(ResultType):
    """Result types with observables and targets.
    If no targets are specified, the observable must only operate on 1 qubit and it
    will be applied to all qubits in parallel. Otherwise, the number of specified targets
    must be equivalent to the number of qubits the observable can be applied to.

    See :mod:`braket.circuits.observables` module for all of the supported observables.
    """

    def __init__(
        self, ascii_symbols: list[str], observable: Observable, target: QubitSetInput | None = None
    ):
        """Initializes an `ObservableResultType`.

        Args:
            ascii_symbols (list[str]): ASCII string symbols for the result type. This is used when
                printing a diagram of circuits.
            observable (Observable): the observable for the result type
            target (QubitSetInput | None): Target qubits that the
                result type is requested for. Default is `None`, which means the observable must
                only operate on 1 qubit and it will be applied to all qubits in parallel

        Raises:
            ValueError: if target=None and the observable's qubit count is not 1.
                Or, if `target!=None` and the observable's qubit count and the number of target
                qubits are not equal. Or, if `target!=None` and the observable's qubit count and
                the number of `ascii_symbols` are not equal.
        """
        super().__init__(ascii_symbols)
        self._observable = observable
        self._target = QubitSet(target)
        if not self._target:
            if self._observable.qubit_count != 1:
                raise ValueError(
                    f"Observable {self._observable} must only operate on 1 qubit for target=None"
                )
        elif isinstance(observable, Sum):  # nested target
            if len(target) != len(observable.summands):
                raise ValueError(
                    "Sum observable's target shape must be a nested list where each term's "
                    "target length is equal to the observable term's qubits count."
                )
            self._target = [QubitSet(term_target) for term_target in target]
            for term_target, obs in zip(target, observable.summands):
                if obs.qubit_count != len(term_target):
                    raise ValueError(
                        "Sum observable's target shape must be a nested list where each term's "
                        "target length is equal to the observable term's qubits count."
                    )
        elif self._observable.qubit_count != len(self._target):
            raise ValueError(
                f"Observable's qubit count {self._observable.qubit_count} and "
                f"the size of the target qubit set {self._target} must be equal"
            )
        elif self._observable.qubit_count != len(self.ascii_symbols):
            raise ValueError(
                "Observable's qubit count and the number of ASCII symbols must be equal"
            )

    @property
    def observable(self) -> Observable:
        return self._observable

    @property
    def target(self) -> QubitSet:
        return self._target

    @target.setter
    def target(self, target: QubitSetInput) -> None:
        """Sets the target.

        Args:
            target (QubitSetInput): The new target.
        """
        self._target = QubitSet(target)

    def __eq__(self, other: ObservableResultType) -> bool:
        if isinstance(other, ObservableResultType):
            return (
                self.name == other.name
                and self.target == other.target
                and self.observable == other.observable
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f"{self.name}(observable={self.observable}, target={self.target})"

    def __copy__(self) -> ObservableResultType:
        return type(self)(observable=self.observable, target=self.target)

    def __hash__(self) -> int:
        return super().__hash__()


class ObservableParameterResultType(ObservableResultType):
    """Result types with observables, targets and parameters.
    If no targets are specified, the observable must only operate on 1 qubit and it
    will be applied to all qubits in parallel. Otherwise, the number of specified targets
    must be equivalent to the number of qubits the observable can be applied to.
    If no parameters are specified, observable will be applied to all the free parameters.

    See :mod:`braket.circuits.observables` module for all of the supported observables.
    """

    def __init__(
        self,
        ascii_symbols: list[str],
        observable: Observable,
        target: QubitSetInput | None = None,
        parameters: list[Union[str, FreeParameter]] | None = None,
    ):
        super().__init__(ascii_symbols, observable, target)

        self._parameters = (
            [(param.name if isinstance(param, FreeParameter) else param) for param in parameters]
            if parameters
            else parameters
        )

        """
        Args:
            ascii_symbols (list[str]): ASCII string symbols for the result type. This is used when
                printing a diagram of circuits.
            observable (Observable): the observable for the result type.
            target (QubitSetInput | None): Target qubits that the result type is requested for.
                Default is `None`, which means the observable must only operate on 1
                qubit and it will be applied to all qubits in parallel.
            parameters (list[Union[str, FreeParameter]] | None): List of string inputs or
                FreeParameter objects. These inputs will be used as parameters for
                gradient calculation. Default: `all`.

        Raises:
            ValueError: if target=None and the observable's qubit count is not 1.
                Or, if `target!=None` and the observable's qubit count and the number of target
                qubits are not equal. Or, if `target!=None` and the observable's qubit count and
                the number of `ascii_symbols` are not equal.
        """

    @property
    def parameters(self) -> list[str]:
        return self._parameters

    def __repr__(self) -> str:
        return (
            f"{self.name}(observable={self.observable}, target={self.target}, "
            f"parameters={self.parameters})"
        )

    def __copy__(self) -> ObservableResultType:
        return type(self)(
            observable=self.observable, target=self.target, parameters=self.parameters
        )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from functools import reduce, singledispatch
from typing import Union

import braket.circuits.gates as braket_gates
import braket.circuits.result_types as ResultTypes  # noqa: N812
import braket.ir.jaqcd.shared_models as models
from braket.circuits import Observable, noises, observables
from braket.ir.jaqcd import (
    Amplitude,
    DensityMatrix,
    Expectation,
    Probability,
    Sample,
    StateVector,
    Variance,
)
from braket.ir.jaqcd.program_v1 import Results

BRAKET_GATES = {
    "gphase": braket_gates.GPhase,
    "i": braket_gates.I,
    "h": braket_gates.H,
    "x": braket_gates.X,
    "y": braket_gates.Y,
    "z": braket_gates.Z,
    "cv": braket_gates.CV,
    "cnot": braket_gates.CNot,
    "cy": braket_gates.CY,
    "cz": braket_gates.CZ,
    "ecr": braket_gates.ECR,
    "s": braket_gates.S,
    "si": braket_gates.Si,
    "t": braket_gates.T,
    "ti": braket_gates.Ti,
    "v": braket_gates.V,
    "vi": braket_gates.Vi,
    "phaseshift": braket_gates.PhaseShift,
    "cphaseshift": braket_gates.CPhaseShift,
    "cphaseshift00": braket_gates.CPhaseShift00,
    "cphaseshift01": braket_gates.CPhaseShift01,
    "cphaseshift10": braket_gates.CPhaseShift10,
    "rx": braket_gates.Rx,
    "ry": braket_gates.Ry,
    "rz": braket_gates.Rz,
    "U": braket_gates.U,
    "swap": braket_gates.Swap,
    "iswap": braket_gates.ISwap,
    "pswap": braket_gates.PSwap,
    "xy": braket_gates.XY,
    "xx": braket_gates.XX,
    "yy": braket_gates.YY,
    "zz": braket_gates.ZZ,
    "ccnot": braket_gates.CCNot,
    "cswap": braket_gates.CSwap,
    "gpi": braket_gates.GPi,
    "gpi2": braket_gates.GPi2,
    "ms": braket_gates.MS,
    "unitary": braket_gates.Unitary,
}

one_prob_noise_map = {
    "bit_flip": noises.BitFlip,
    "phase_flip": noises.PhaseFlip,
    "pauli_channel": noises.PauliChannel,
    "depolarizing": noises.Depolarizing,
    "two_qubit_depolarizing": noises.TwoQubitDepolarizing,
    "two_qubit_dephasing": noises.TwoQubitDephasing,
    "amplitude_damping": noises.AmplitudeDamping,
    "generalized_amplitude_damping": noises.GeneralizedAmplitudeDamping,
    "phase_damping": noises.PhaseDamping,
}

SUPPORTED_NOISE_PRAGMA_TO_NOISE = {
    "braket_noise_bit_flip": noises.BitFlip,
    "braket_noise_phase_flip": noises.PhaseFlip,
    "braket_noise_pauli_channel": noises.PauliChannel,
    "braket_noise_depolarizing": noises.Depolarizing,
    "braket_noise_two_qubit_depolarizing": noises.TwoQubitDepolarizing,
    "braket_noise_two_qubit_dephasing": noises.TwoQubitDephasing,
    "braket_noise_amplitude_damping": noises.AmplitudeDamping,
    "braket_noise_generalized_amplitude_damping": noises.GeneralizedAmplitudeDamping,
    "braket_noise_phase_damping": noises.PhaseDamping,
    "braket_noise_kraus": noises.Kraus,
}


def get_observable(obs: Union[models.Observable, list]) -> Observable:
    """Gets the observable.

    Args:
        obs (Union[Observable, list]): The observable(s) to get translated.

    Returns:
        Observable: The translated observable.
    """
    return _get_observable(obs)


@singledispatch
def _get_observable(obs: Union[models.Observable, list]) -> Observable:
    raise NotImplementedError


@_get_observable.register(list)
def _(obs):
    raise NotImplementedError


@_get_observable.register(str)
def _(name: str):
    return getattr(observables, name.upper())()


def get_tensor_product(observable: Union[models.Observable, list]) -> Observable:
    """Generate an braket circuit observable

    Args:
        observable (Union[Observable, list]): ir observable or a matrix

    Returns:
        Observable: braket circuit observable
    """
    circuit_observable = [get_observable(obs) for obs in observable]
    return reduce(lambda obs1, obs2: obs1 @ obs2, circuit_observable)


@singledispatch
def _braket_result_to_result_type(result: Results) -> None:
    raise TypeError(f"Result type {type(result).__name__} is not supported")


def braket_result_to_result_type(result: Results) -> None:
    return _braket_result_to_result_type(result)


@_braket_result_to_result_type.register(Amplitude)
def _(result: Results) -> Amplitude:
    return ResultTypes.Amplitude(state=result.states)


@_braket_result_to_result_type.register(Expectation)
def _(result: Results) -> Expectation:
    tensor_product = get_tensor_product(result.observable)

    return ResultTypes.Expectation(observable=tensor_product, target=result.targets)


@_braket_result_to_result_type.register(Probability)
def _(result: Results) -> Probability:
    return ResultTypes.Probability(result.targets)


@_braket_result_to_result_type.register(Sample)
def _(result: Results) -> Sample:
    tensor_product = get_tensor_product(result.observable)
    return ResultTypes.Sample(observable=tensor_product, target=result.targets)


@_braket_result_to_result_type.register(StateVector)
def _(result: Results) -> StateVector:
    return ResultTypes.StateVector()


@_braket_result_to_result_type.register(DensityMatrix)
def _(result: Results):
    return ResultTypes.DensityMatrix(target=result.targets)


@_braket_result_to_result_type.register(Variance)
def _(result: Results):
    tensor_product = get_tensor_product(result.observable)
    return ResultTypes.Variance(observable=tensor_product, target=result.targets)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from braket.jobs.metrics_data.definitions import MetricStatistic, MetricType


class QuantumJob(ABC):
    DEFAULT_RESULTS_POLL_TIMEOUT = 864000
    DEFAULT_RESULTS_POLL_INTERVAL = 5

    @property
    @abstractmethod
    def arn(self) -> str:
        """The ARN (Amazon Resource Name) of the hybrid job.

        Returns:
            str: The ARN (Amazon Resource Name) of the hybrid job.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the hybrid job.

        Returns:
            str: The name of the hybrid job.
        """

    @abstractmethod
    def state(self, use_cached_value: bool = False) -> str:
        """The state of the hybrid job.

        Args:
            use_cached_value (bool): If `True`, uses the value most recently retrieved
                value from the Amazon Braket `GetJob` operation. If `False`, calls the
                `GetJob` operation to retrieve metadata, which also updates the cached
                value. Default = `False`.

        Returns:
            str: The value of `status` in `metadata()`. This is the value of the `status` key
            in the Amazon Braket `GetJob` operation.

        See Also:
            `metadata()`
        """

    @abstractmethod
    def logs(self, wait: bool = False, poll_interval_seconds: int = 5) -> None:
        """Display logs for a given hybrid job, optionally tailing them until hybrid job is
        complete.

        If the output is a tty or a Jupyter cell, it will be color-coded
        based on which instance the log entry is from.

        Args:
            wait (bool): `True` to keep looking for new log entries until the hybrid job completes;
                otherwise `False`. Default: `False`.

            poll_interval_seconds (int): The interval of time, in seconds, between polling for
                new log entries and hybrid job completion (default: 5).

        Raises:
            RuntimeError: If waiting and the hybrid job fails.
        """
        # The loop below implements a state machine that alternates between checking the hybrid job
        # status and reading whatever is available in the logs at this point. Note, that if we were
        # called with wait == False, we never check the hybrid job status.
        #
        # If wait == TRUE and hybrid job is not completed, the initial state is TAILING
        # If wait == FALSE, the initial state is COMPLETE (doesn't matter if the hybrid job really
        # is complete).
        #
        # The state table:
        #
        # STATE               ACTIONS                        CONDITION             NEW STATE
        # ----------------    ----------------               -----------------     ----------------
        # TAILING             Read logs, Pause, Get status   Job complete          JOB_COMPLETE
        #                                                    Else                  TAILING
        # JOB_COMPLETE        Read logs, Pause               Any                   COMPLETE
        # COMPLETE            Read logs, Exit                                      N/A
        #
        # Notes:
        # - The JOB_COMPLETE state forces us to do an extra pause and read any items that got to
        #   Cloudwatch after the job was marked complete.

    @abstractmethod
    def metadata(self, use_cached_value: bool = False) -> dict[str, Any]:
        """Gets the job metadata defined in Amazon Braket.

        Args:
            use_cached_value (bool): If `True`, uses the value most recently retrieved
                from the Amazon Braket `GetJob` operation, if it exists; if does not exist,
                `GetJob` is called to retrieve the metadata. If `False`, always calls
                `GetJob`, which also updates the cached value. Default: `False`.

        Returns:
            dict[str, Any]: Dict that specifies the hybrid job metadata defined in Amazon Braket.
        """

    @abstractmethod
    def metrics(
        self,
        metric_type: MetricType = MetricType.TIMESTAMP,
        statistic: MetricStatistic = MetricStatistic.MAX,
    ) -> dict[str, list[Any]]:
        """Gets all the metrics data, where the keys are the column names, and the values are a list
        containing the values in each row.

        Args:
            metric_type (MetricType): The type of metrics to get. Default: MetricType.TIMESTAMP.

            statistic (MetricStatistic): The statistic to determine which metric value to use
                when there is a conflict. Default: MetricStatistic.MAX.

        Returns:
            dict[str, list[Any]]: The metrics data.

        Example:
            timestamp energy
              0         0.1
              1         0.2
            would be represented as:
            { "timestamp" : [0, 1], "energy" : [0.1, 0.2] }
            values may be integers, floats, strings or None.
        """

    @abstractmethod
    def cancel(self) -> str:
        """Cancels the hybrid job.

        Returns:
            str: Indicates the status of the hybrid job.

        Raises:
            ClientError: If there are errors invoking the CancelJob API.
        """

    @abstractmethod
    def result(
        self,
        poll_timeout_seconds: float = DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: float = DEFAULT_RESULTS_POLL_INTERVAL,
    ) -> dict[str, Any]:
        """Retrieves the hybrid job result persisted using save_job_result() function.

        Args:
            poll_timeout_seconds (float): The polling timeout, in seconds, for `result()`.
                Default: 10 days.

            poll_interval_seconds (float): The polling interval, in seconds, for `result()`.
                Default: 5 seconds.


        Returns:
            dict[str, Any]: Dict specifying the hybrid job results.

        Raises:
            RuntimeError: if hybrid job is in a FAILED or CANCELLED state.
            TimeoutError: if hybrid job execution exceeds the polling timeout period.
        """

    @abstractmethod
    def download_result(
        self,
        extract_to: str | None = None,
        poll_timeout_seconds: float = DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: float = DEFAULT_RESULTS_POLL_INTERVAL,
    ) -> None:
        """Downloads the results from the hybrid job output S3 bucket and extracts the tar.gz
        bundle to the location specified by `extract_to`. If no location is specified,
        the results are extracted to the current directory.

        Args:
            extract_to (str | None): The directory to which the results are extracted. The results
                are extracted to a folder titled with the hybrid job name within this directory.
                Default= `Current working directory`.

            poll_timeout_seconds (float): The polling timeout, in seconds, for `download_result()`.
                Default: 10 days.

            poll_interval_seconds (float): The polling interval, in seconds, for
                `download_result()`.Default: 5 seconds.

        Raises:
            RuntimeError: if hybrid job is in a FAILED or CANCELLED state.
            TimeoutError: if hybrid job execution exceeds the polling timeout period.
        """
